(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import finite.

Require Import good_base.

Require Import af.
Require Import bar.
Require Import good_bar.
Require Import af_bar.
Require Import fan.

Require Import af_t.
Require Import bar_t.
Require Import good_bar_t.
Require Import af_bar_t.
Require Import fan_t.

Set Implicit Arguments.

Section abtree.

  (* at most binary trees *)

  Variable X Y Z : Type.

  Inductive abtree : Type :=
    | abt_leaf : X -> abtree
    | abt_cons : Y -> abtree -> abtree
    | abt_pair : Z -> abtree -> abtree -> abtree.

  Inductive sub_abtree : abtree -> abtree -> Prop :=
    | in_sabt_0 : forall t, sub_abtree t t
    | in_sabt_1 : forall s y t, sub_abtree s t -> sub_abtree s (abt_cons y t)
    | in_sabt_2 : forall s z t1 t2, sub_abtree s t1 -> sub_abtree s (abt_pair z t1 t2)
    | in_sabt_3 : forall s z t1 t2, sub_abtree s t2 -> sub_abtree s (abt_pair z t1 t2).

  Fact sub_abtree_inv s t : sub_abtree s t -> s = t 
                                           \/ (exists y t', sub_abtree s t' /\ t = abt_cons y t')
                                           \/ (exists z t1 t2, (sub_abtree s t1 \/ sub_abtree s t2) /\ t = abt_pair z t1 t2).
  Proof.
    inversion_clear 1.
    left; auto.
    right; left; exists y, t0; auto.
    right; right; exists z, t1, t2; tauto.
    right; right; exists z, t1, t2; tauto.
  Qed.

  Fixpoint abt_size t :=
    match t with 
      | abt_leaf   x   => 0
      | abt_cons _ t   => S (abt_size t)
      | abt_pair _ t1 t2 => S (abt_size t1 + abt_size t2)
    end.

  Fact sub_abtree_size s t : sub_abtree s t -> abt_size s <= abt_size t.
  Proof.
    induction 1; simpl; omega.
  Qed.

  Fact sub_abtree_size_eq s t : sub_abtree s t -> (abt_size s = abt_size t <-> s = t).
  Proof.
    induction 1 as [ | s y t H1 IH | s z t1 t2 H1 IH | s z t1 t2 H1 IH ]; simpl; split; auto;
    ( apply sub_abtree_size in H1; try omega; fail || intros; subst; simpl; auto ).
  Qed.

  Fact sub_abtree_leaf_inv t x : sub_abtree t (abt_leaf x) -> t = abt_leaf x.
  Proof.     
    inversion_clear 1; auto. 
  Qed.
  
  Fact sub_abtree_leaf_cons_inv x y t : sub_abtree (abt_leaf x) (abt_cons y t) -> sub_abtree (abt_leaf x) t.
  Proof.
    inversion_clear 1; auto.
  Qed.
  
  Fact sub_abtree_leaf_pair_inv x z t1 t2 : sub_abtree (abt_leaf x) (abt_pair z t1 t2) -> sub_abtree (abt_leaf x) t1 \/ sub_abtree (abt_leaf x) t2.
  Proof.
    inversion_clear 1; auto.
  Qed.

  Fact sub_abtree_cons_cons_inv x s y t : sub_abtree (abt_cons x s) (abt_cons y t) -> x = y /\ s = t \/ sub_abtree (abt_cons x s) t.
  Proof.
    inversion_clear 1; auto.
  Qed.

  Fact sub_abtree_cons_pair_inv y s z t1 t2 : sub_abtree (abt_cons y s) (abt_pair z t1 t2) -> sub_abtree (abt_cons y s) t1
                                                                                           \/ sub_abtree (abt_cons y s) t2.
  Proof.
    inversion_clear 1; auto.
  Qed.

  Fact sub_abtree_cons_inv s y t : sub_abtree s (abt_cons y t) -> s = abt_cons y t \/ sub_abtree s t. 
  Proof.
    intros H.
    apply sub_abtree_inv in H.
    destruct H as [ H | [ (y0 & t' & H1 & H2) | (z & t1 & t2 & [ H1 | H1 ] & H2) ] ]; subst; auto; try discriminate H2.
    injection H2; intros; subst; auto.
  Qed.

  Fact sub_abtree_pair_inv s z t1 t2 : sub_abtree s (abt_pair z t1 t2) -> s = abt_pair z t1 t2
                                                                       \/ sub_abtree s t1
                                                                       \/ sub_abtree s t2.
  Proof.
    intros H.
    apply sub_abtree_inv in H.
    destruct H as [ H | [ (y0 & t' & H1 & H2) | (z0 & t1' & t2' & [ H1 | H1 ] & H2) ] ]; subst; auto; try discriminate H2;
    injection H2; intros; subst; auto.    
  Qed.

  Variables (RX : X -> X -> Prop) (RY : Y -> Y -> Prop) (RZ : Z -> Z -> Prop).

  Reserved Notation "x '<t' y" (at level 70, no associativity).

  Inductive embed_abtree : abtree -> abtree -> Prop :=
    | in_eabt_0 : forall x y,             RX x y ->                               abt_leaf x <t abt_leaf y
    | in_eabt_1 : forall x s y t,         RY x y ->  s <t t ->                  abt_cons x s <t abt_cons y t
    | in_eabt_2 : forall x s1 s2 y t1 t2, RZ x y -> s1 <t t1 -> s2 <t t2 -> abt_pair x s1 s2 <t abt_pair y t1 t2
    | in_eabt_3 : forall s x t,                      s <t t ->                             s <t abt_cons x t
    | in_eabt_4 : forall s x t1 t2,                  s <t t1 ->                            s <t abt_pair x t1 t2
    | in_eabt_5 : forall s x t1 t2,                  s <t t2 ->                            s <t abt_pair x t1 t2
  where "x <t y" := (embed_abtree x y).

  Fact sabt_leaf_leaf_inv x y : abt_leaf x <t abt_leaf y -> RX x y.
  Proof. inversion_clear 1; auto. Qed.

  Fact sabt_leaf_cons_inv x y t : abt_leaf x <t abt_cons y t -> abt_leaf x <t t.
  Proof. inversion_clear 1; auto. Qed.
  
  Fact sabt_leaf_pair_inv x y t1 t2 : abt_leaf x <t abt_pair y t1 t2 -> abt_leaf x <t t1 \/ abt_leaf x <t t2.
  Proof. inversion_clear 1; auto. Qed.
  
  Fact sabt_all_leaf_inv t y : t <t abt_leaf y -> exists x, RX x y /\ t = abt_leaf x.
  Proof. inversion_clear 1; exists x; auto. Qed.

  Fixpoint abt_leaves t := 
    match t with 
      | abt_leaf   x   => x::nil
      | abt_cons _ t   => abt_leaves t
      | abt_pair _ s t => abt_leaves s ++ abt_leaves t
    end.

  Fact abt_leaf_exists t : { x | In x (abt_leaves t) }.
  Proof.
    induction t as [ x | y ? (x & Hx) | z ? (x & Hx) ? _ ].
    exists x; simpl; auto.
    exists x; simpl; auto.
    exists x; simpl; apply in_or_app; left; auto.
  Defined.

  Fact sabt_leaf_inv x t : abt_leaf x <t t <-> exists y, In y (abt_leaves t) /\ RX x y.
  Proof.
    split.

    induction t as [ y | y t Ht | y t1 Ht1 t2 Ht2 ]; intros H.
    apply sabt_leaf_leaf_inv in H; exists y; simpl; auto.
    apply sabt_leaf_cons_inv in H; destruct (Ht H) as (z & H1 & H2); exists z; simpl; auto.
    apply sabt_leaf_pair_inv in H; destruct H as [ H | H ].
    destruct (Ht1 H) as (z & H1 & H2); exists z; simpl; split; auto; apply in_or_app; tauto.
    destruct (Ht2 H) as (z & H1 & H2); exists z; simpl; split; auto; apply in_or_app; tauto.
    
    intros (y & H1 & H2); revert H1 x H2.
    induction t as [ z | z t Ht | z t1 Ht1 t2 Ht2 ]; simpl.
    intros [ ? | [] ] x Hx; subst; constructor 1; auto.
    constructor 4; auto.
    intros H; apply in_app_or in H; destruct H; [ constructor 5 | constructor 6 ]; auto.
  Qed.
  
  Fact sabt_cons_leaf_inv x s y : ~ abt_cons x s <t abt_leaf y.
  Proof. red; inversion 1. Qed.
  
  Fact sabt_cons_cons_inv x s y t : abt_cons x s <t abt_cons y t -> RY x y /\ s <t t \/ abt_cons x s <t t.
  Proof. inversion_clear 1; tauto. Qed.
  
  Fact sabt_cons_pair_inv x s y t1 t2 : abt_cons x s <t abt_pair y t1 t2 -> abt_cons x s <t t1 \/ abt_cons x s <t t2.
  Proof. inversion_clear 1; tauto. Qed.

  Fact sabt_pair_leaf_inv x s1 s2 y : ~ abt_pair x s1 s2 <t abt_leaf y.
  Proof. red; inversion 1. Qed.
  
  Fact sabt_pair_cons_inv x s1 s2 y t1 : abt_pair x s1 s2 <t abt_cons y t1 -> abt_pair x s1 s2 <t t1. 
  Proof. inversion_clear 1; tauto. Qed.
  
  Fact sabt_pair_pair_inv x s1 s2 y t1 t2 : abt_pair x s1 s2 <t abt_pair y t1 t2 
                                         -> RZ x y /\ s1 <t t1 /\ s2 <t t2 
                                         \/ abt_pair x s1 s2 <t t1
                                         \/ abt_pair x s1 s2 <t t2.
  Proof. inversion_clear 1; tauto. Qed.

  Fact embed_sub_abtree r s t : embed_abtree r s -> sub_abtree s t -> embed_abtree r t.
  Proof.
    induction 2; auto.
    constructor 4; auto.
    constructor 5; auto.
    constructor 6; auto.
  Qed.

  Fact sub_embed_abtree r s t : sub_abtree r s -> embed_abtree s t -> embed_abtree r t.
  Proof.
    intros H.
    induction 1.

    apply sub_abtree_leaf_inv in H; subst; constructor 1; auto.

    apply sub_abtree_cons_inv in H; destruct H; subst.
    constructor 2; auto.
    constructor 4; auto.
    
    apply sub_abtree_pair_inv in H. 
    destruct H as [ | [|] ]; subst.
    constructor 3; auto.
    constructor 5; auto.
    constructor 6; auto.
    
    constructor 4; auto.

    constructor 5; auto.
    
    constructor 6; auto.
  Qed.

End abtree.

Implicit Arguments abt_leaf [ X Y Z ].
Implicit Arguments abt_cons [ X Y Z ].
Implicit Arguments abt_pair [ X Y Z ].

Fact abtree_no_leaf Y Z (t : abtree ES Y Z) : False.
Proof.
  destruct (abt_leaf_exists t) as ( [] & _) .
Qed.

Section embed_abtree_mono.

  Variable X Y Z : Type.
  Variables (RX1 RX2 : X -> X -> Prop) (RY1 RY2 : Y -> Y -> Prop) (RZ1 RZ2 : Z -> Z -> Prop).
  
  Fact embed_autree_mono : RX1 inc2 RX2 -> RY1 inc2 RY2 -> RZ1 inc2 RZ2 
                        -> embed_abtree RX1 RY1 RZ1 inc2 embed_abtree RX2 RY2 RZ2.
  Proof.
    intros HX HY HZ.
    induction 1.
    constructor 1; auto.
    constructor 2; auto.
    constructor 3; auto.
    constructor 4; auto.
    constructor 5; auto.
    constructor 6; auto.
  Qed.

End embed_abtree_mono.

Section abtree_ev1.

  Variable (X Y Z : Type).

  Let X' := (X + Y * abtree X Y Z)%type.
  
  Let ev := fix loop (t : abtree X' Y Z) : abtree X Y Z :=
    match t with
      | abt_leaf (inl x)      => abt_leaf x 
      | abt_leaf (inr (y,s))  => abt_cons y s
      | abt_cons y t1         => abt_cons y (loop t1)
      | abt_pair z t1 t2      => abt_pair z (loop t1) (loop t2)
    end.

  Let ev_inv := fix loop (t : abtree X Y Z) : list (abtree X' Y Z) :=
    match t with
      | abt_leaf x          => abt_leaf (inl x)::nil
      | abt_cons y t1       => abt_leaf (inr (y,t1))::map (abt_cons y) (loop t1)
      | abt_pair z t1 t2    => list_prod (abt_pair z) (loop t1) (loop t2) 
    end.

  Let ev_inv_spec s t : In t (ev_inv s) <-> s = ev t.
  Proof.
    revert t; induction s as [ x | y s1 Hs1 | z s1 Hs1 s2 Hs2 ]; intros t; split; simpl.

    intros [ ? | [] ]; subst; auto.
    destruct t as [ [|(?,?)] | | ]; simpl; try discriminate 1.
    injection 1; intros; subst; auto.
    
    intros [ H | H ].
    subst; auto.
    rewrite in_map_iff in H.
    destruct H as (t1 & ? & Ht1); subst; simpl.
    f_equal; auto.
    apply Hs1; auto.
    
    destruct t as [ [|(?,?)] | y' t1 | ]; simpl; try discriminate 1.
    intros H; injection H; intros ? ?; subst; tauto.
    intros H; injection H; intros ? ?; subst.
    right.
    apply in_map_iff.
    exists t1; split; simpl.
    f_equal; auto.
    apply Hs1; auto.
    
    intros H.
    apply list_prod_spec in H.
    destruct H as (t1 & t2 & H1 & H2 & ?); subst; simpl.
    f_equal; auto.
    apply Hs1; auto.
    apply Hs2; auto.
    
    destruct t as [ [|(?,?)] | | z' t1 t2 ]; simpl; try discriminate 1.
    intros H; injection H; intros ? ? ?; subst.
    apply list_prod_spec.
    exists t1, t2; repeat split.
    apply Hs1; auto.
    apply Hs2; auto.
  Qed.

  Notation analysis := (fun s a => s = ev a).
  
  Let analysis_finite s : finite (analysis s).
  Proof.
    exists (ev_inv s).
    intros; apply ev_inv_spec.
  Qed.
  
  Variables (RX : X -> X -> Prop) (RY : Y -> Y -> Prop) (RZ : Z -> Z -> Prop).

  Variables (m : Y) (t : abtree X Y Z).

  Definition abt_rel1 := sum_both RX (rel_prod RY (embed_abtree RX RY RZ rlift t)).
  
  Fact abt_rel1_af : af RX -> af RY -> af (embed_abtree RX RY RZ rlift t) -> af abt_rel1.
  Proof.
    intros H1 H2 H3.
    apply af_sum_both; auto.
    apply af_product; auto.
  Qed. 

  Fact abt_rel1_af_t : af_t RX -> af_t RY -> af_t (embed_abtree RX RY RZ rlift t) -> af_t abt_rel1.
  Proof.
    intros H1 H2 H3.
    apply af_t_sum_both; auto.
    apply af_t_prod; auto.
  Qed. 

  Notation SX := abt_rel1.
  Notation SY := (RY rlift m).

  Let disapointing (s : abtree X' Y Z) := 
    match s with
      | abt_pair _ _ _       => False
      | abt_cons y _         => RY m y
      | abt_leaf (inl _)     => False
      | abt_leaf (inr (_,v)) => embed_abtree RX RY RZ t v
    end.
  
  Let has_disap s := exists a, disapointing a /\ sub_abtree a s.

  Let disap_prop s : analysis s inc1 has_disap
                  -> embed_abtree RX RY RZ (abt_cons m t) s.
  Proof.
    induction s as [ x | y s IHs | z t1 IHt1 t2 IHt2 ]; simpl; intros H.

    destruct (H (abt_leaf (inl x))) as (a & H1 & H2).
    auto.
    inversion H2; subst.
    destruct H1.
    
    destruct (H (abt_leaf (inr (y,s)))) as (a & H1 & H2).
    tauto.
    apply sub_abtree_leaf_inv in H2; subst a.
    simpl in H1.

    (* we decide on the analysis of s *) 
 
    destruct (@finite_decide _ (analysis s)) 
      with (R := has_disap)
           (S := fun a => disapointing (abt_cons y a))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.

    intros x Hx.
    destruct (H (abt_cons y x)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_cons_inv in Hb2; 
      destruct Hb2 as [ | Hb2 ].
    subst b; right; auto.
    left; exists b; auto.
    
    constructor 4; apply IHs; auto.
    red in Hb2; subst s; constructor 2; auto.
    
    (* we decide on the product of analysis of both t1 and t2 *)
    
    destruct (@finite_decide _ (pred_prod (analysis t1) (analysis t2)))
      with (R := fun c => has_disap (fst c) \/ has_disap (snd c))
           (S := fun c => disapointing (abt_pair z (fst c) (snd c)))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.
    apply finite_product; auto.
    
    intros (u1,u2) (H1 & H2); simpl in H1, H2.
    destruct (H (abt_pair z u1 u2)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_pair_inv in Hb2. 
      destruct Hb2 as [ | [ Hb2 | Hb2 ] ].
    subst b; right; auto.
    left; left; exists b; auto.
    left; right; exists b; auto.
    
    apply product_analysis in H0; auto.
    destruct H0 as [ H0 | H0 ].
    constructor 5; auto.
    constructor 6; auto.
    
    simpl in Hb2; destruct Hb2.
  Qed.

  Let embed_prop s1 s2 : embed_abtree SX SY RZ s1 s2 
                      -> embed_abtree RX RY RZ (ev s1) (ev s2)
                      \/ has_disap s1.
  Proof.
    induction 1 as [ [ x1 | (y1,s1) ] [ x2 | (y2,s2) ] H1 
                   | y1 s1 y2 s2 H1 H2 [ IH | (d & Hd1 & Hd2) ]
                   | z1 s1 t1 z2 s2 t2 H0 H1 [ IH1 | (d1 & Hd1 & Hd2) ] H2 [ IH2 | (d2 & Hd3 & Hd4) ]
                   | s1 y2 s2 H1 [ IH | (d & Hd1 & Hd2) ]
                   | s1 z1 t1 t2 H1 [ IH1 | (d & Hd1 & Hd2) ]
                   | s1 z1 t1 t2 H2 [ IH2 | (d & Hd1 & Hd2) ]
                   ]; simpl in * |- *; try (destruct H1; fail).

    left; constructor 1; auto.
    red in H1; simpl in H1; destruct H1 as [ H1 H2 ].
    destruct H2 as [ H2 | H2 ].
    left; constructor 2; auto.
    right; exists (abt_leaf (inr (y1,s1))); simpl; split; auto; constructor 1; auto.
    
    destruct H1.
    left; constructor 2; auto.
    right; exists (abt_cons y1 s1); simpl; split; auto; constructor 1.    
    right; exists d; split; auto; constructor 2; auto.
    
    left; constructor 3; auto.
    right; exists d2; split; auto; constructor 4; auto.
    right; exists d1; split; auto; constructor 3; auto.
    right; exists d1; split; auto; constructor 3; auto.
    
    left; constructor 4; auto.
    right; exists d; auto.
    
    left; constructor 5; auto.
    right; exists d; auto.
    
    left; constructor 6; auto.
    right; exists d; auto.
  Qed.

  Let good_or_disap l : 
                   good (embed_abtree SX SY RZ) l
                -> good (embed_abtree RX RY RZ) (map ev l)
                \/ exists x, In x l /\ has_disap x.
  Proof.
    induction 1 as [ l x1 x2 H1 H2 
                   | l x1 x2 [ H1 | (x & H1 & H2) ] ].
                   
    apply embed_prop in H2.
    destruct H2 as [ H2 | H2 ].
    left; simpl; constructor 1 with (ev x2); auto.
    rewrite in_map_iff; exists x2; auto.
    right; exists x2; simpl; tauto.
    
    left; simpl; constructor 2; auto.
    
    right; exists x; simpl; tauto.
  Qed.

  Notation R' := (embed_abtree SX SY RZ).

  Hypothesis HST : af R'.
  Hypothesis HST_t : af_t R'.
  
  Let HST' : bar (good R') nil.
  Proof.
    revert HST; apply af_eq_bar_good.
  Qed.
  
  Let HST_t' : bar_t (good R') nil.
  Proof.
    revert HST_t; apply af_t_bar_t.
  Qed.

  Notation U := (fun ws => Forall (good R') (list_fan (map ev_inv ws))).
  
  Let HU : bar U nil.
  Proof.
    apply bar_map_inv with (f := ev_inv) 
                           (Q := fun ll => Forall (good R') (list_fan ll)); auto.
    simpl; apply fan_on_list; auto.
    constructor; auto.
  Qed.
  
  Let HU_t : bar_t U nil.
  Proof.
    apply bar_t_map_inv with (f := ev_inv) 
                           (Q := fun ll => Forall (good R') (list_fan ll)); auto.
    simpl; apply fan_t_on_list; auto.
    constructor; auto.
  Qed.
  
  Let U_to_good ws : U ws -> good (embed_abtree RX RY RZ) (ws ++ (abt_cons m t) :: nil).
  Proof.
    set (Q l := good (embed_abtree RX RY RZ) (map ev l)).
    set (B a := has_disap a).
    
    intros H0.
    destruct (@hig_combi_principle _ (map ev_inv ws) Q B) as [ (p & Q1 & Q2) | (p & Q1 & Q2) ].
    
    (* condition of hig_combi_principle *)
    
    intros l Hl.
    apply list_fan_spec in Hl.
    simpl in H0; rewrite Forall_forall in H0.
    apply H0 in Hl; revert Hl.
    apply good_or_disap; auto.
    
    (* there is a good sequence p that evaluates to ws *)
      
    unfold Q in Q2.
    apply good_app_right.
    replace_with Q2.
    symmetry; apply Forall2_eq.
    revert Q1.
    rewrite Forall2_map_right, Forall2_map_left.
    apply Forall2_impl.
    intros ? ?; rewrite ev_inv_spec; auto.
    
    (* there is a fan in which every element contains a dispointing subtree *)
    
    unfold B in Q2.
    rewrite in_map_iff in Q1.
    destruct Q1 as (x & ? & Q1); subst p.
    apply in_split in Q1.
    destruct Q1 as (l1 & l2 & Q1).
    subst ws.
    rewrite <- app_assoc; simpl.
    apply good_app_left.
    constructor 1 with (abt_cons m t).
    apply in_or_app; right; left; auto.
    apply disap_prop.
    intros ?; rewrite <- ev_inv_spec; auto.
  Qed.

  Let bar_U_bar_good ws : bar U ws -> bar (good (embed_abtree RX RY RZ rlift abt_cons m t)) ws.
  Proof.
    induction 1.
    apply bar_lift_rel, in_bar_0; auto.
    apply in_bar_1; auto.
  Qed.  
  
  Let bar_t_U_bar_t_good ws : bar_t U ws -> bar_t (good (embed_abtree RX RY RZ rlift abt_cons m t)) ws.
  Proof.
    induction 1 using bar_t_rect.
    apply bar_t_lift_rel_rev, in_bar_t0; auto.
    apply in_bar_t1; auto.
  Qed.  

  Theorem abt_ev1_af : af (embed_abtree RX RY RZ rlift abt_cons m t).
  Proof. 
    apply af_eq_bar_good; auto.
  Qed. 
   
  Theorem abt_ev1_af_t : af_t (embed_abtree RX RY RZ rlift abt_cons m t).
  Proof. 
    apply bar_t_af_t; auto.
  Qed. 
  
End abtree_ev1.

Section abtree_ev2.

  Variable (X Y Z : Type).

  Let X' := (X + Y * abtree X Y Z)%type.

  Let ev := fix loop (t : abtree X' ES Z) : abtree X Y Z :=
    match t with
      | abt_leaf (inl x)      => abt_leaf x 
      | abt_leaf (inr (y,s))  => abt_cons y s
      | abt_cons y _          => Empty_set_rect _ y
      | abt_pair z t1 t2      => abt_pair z (loop t1) (loop t2)
    end.

  Let ev_inv := fix loop (t : abtree X Y Z) : list (abtree X' ES Z) :=
    match t with
      | abt_leaf x          => abt_leaf (inl x)::nil
      | abt_cons y t1       => abt_leaf (inr (y,t1))::nil
      | abt_pair z t1 t2    => list_prod (abt_pair z) (loop t1) (loop t2) 
    end.

  Fact ev_inv_spec s t : In t (ev_inv s) <-> s = ev t.
  Proof.
    revert t; induction s as [ x | y s1 Hs1 | z s1 Hs1 s2 Hs2 ]; intros t; split; simpl.

    intros [ ? | [] ]; subst; auto.
    destruct t as [ [|(?,?)] | [] | ]; simpl; try discriminate 1.
    injection 1; intros; subst; auto.
    
    intros [ H | [] ].
    subst; auto.
    left.
    destruct t as [ [|(?,?)] | [] | ]; simpl; try discriminate H.
    simpl in H.
    injection H; intros; subst; auto.
    
    intros H.
    apply list_prod_spec in H.
    destruct H as (t1 & t2 & H1 & H2 & ?); subst; simpl.
    f_equal; auto.
    apply Hs1; auto.
    apply Hs2; auto.
    
    destruct t as [ [|(?,?)] | [] | z' t1 t2 ]; simpl; try discriminate 1.
    intros H; injection H; intros ? ? ?; subst.
    apply list_prod_spec.
    exists t1, t2; repeat split.
    apply Hs1; auto.
    apply Hs2; auto.
  Qed.

  Notation analysis := (fun s a => s = ev a).
  
  Let analysis_finite s : finite (analysis s).
  Proof.
    exists (ev_inv s).
    intros; apply ev_inv_spec.
  Qed.

  Variables (RX : X -> X -> Prop) (RY : Y -> Y -> Prop) (RZ : Z -> Z -> Prop).

  Variables (m : Y) (t : abtree X Y Z).

  Definition abt_rel2 := sum_both RX (rel_prod RY (embed_abtree RX RY RZ rlift t)).
  
  Fact abt_rel2_af : af RX -> full RY -> af (embed_abtree RX RY RZ rlift t) -> af abt_rel2.
  Proof.
    intros H1 H2 H3.
    apply af_sum_both; auto.
    apply af_product; auto.
    constructor 1; auto.
  Qed. 

  Fact abt_rel2_af_t : af_t RX -> full RY -> af_t (embed_abtree RX RY RZ rlift t) -> af_t abt_rel2.
  Proof.
    intros H1 H2 H3.
    apply af_t_sum_both; auto.
    apply af_t_prod; auto.
    apply in_af_t0; auto.
  Qed. 

  Notation SX := abt_rel2.

  Let disapointing (s : abtree X' ES Z) := 
    match s with
      | abt_pair _ _ _       => False
      | abt_cons _ _         => False
      | abt_leaf (inl _)     => False
      | abt_leaf (inr (_,v)) => embed_abtree RX RY RZ t v
    end.
  
  Let has_disap s := exists a, disapointing a /\ sub_abtree a s.

  Hypothesis HRY : full RY.

  Let disap_prop s : analysis s inc1 has_disap
                  -> embed_abtree RX RY RZ (abt_cons m t) s.
  Proof.
    induction s as [ x | y s IHs | z t1 IHt1 t2 IHt2 ]; simpl; intros H.

    destruct (H (abt_leaf (inl x))) as (a & H1 & H2).
    auto.
    inversion H2; subst.
    destruct H1.
    
    destruct (H (abt_leaf (inr (y,s)))) as (a & H1 & H2).
    tauto.
    apply sub_abtree_leaf_inv in H2; subst a.
    simpl in H1.
    
    constructor 2; auto.

    (* we decide on the product of analysis of both t1 and t2 *)
    
    destruct (@finite_decide _ (pred_prod (analysis t1) (analysis t2)))
      with (R := fun c => has_disap (fst c) \/ has_disap (snd c))
           (S := fun c => disapointing (abt_pair z (fst c) (snd c)))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.
    apply finite_product; auto.
    
    intros (u1,u2) (H1 & H2); simpl in H1, H2.
    destruct (H (abt_pair z u1 u2)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_pair_inv in Hb2. 
      destruct Hb2 as [ | [ Hb2 | Hb2 ] ].
    subst b; right; auto.
    left; left; exists b; auto.
    left; right; exists b; auto.
    
    apply product_analysis in H0; auto.
    destruct H0 as [ H0 | H0 ].
    constructor 5; auto.
    constructor 6; auto.
    
    simpl in Hb2; destruct Hb2.
  Qed.

  Let embed_prop s1 s2 : embed_abtree SX ER RZ s1 s2 
                      -> embed_abtree RX RY RZ (ev s1) (ev s2)
                      \/ has_disap s1.
  Proof.
    induction 1 as [ [ x1 | (y1,s1) ] [ x2 | (y2,s2) ] H1 
                   | [] s1 y2 s2 H1 H2 [ IH | (d & Hd1 & Hd2) ]
                   | z1 s1 t1 z2 s2 t2 H0 H1 [ IH1 | (d1 & Hd1 & Hd2) ] H2 [ IH2 | (d2 & Hd3 & Hd4) ]
                   | s1 [] s2 H1 [ IH | (d & Hd1 & Hd2) ]
                   | s1 z1 t1 t2 H1 [ IH1 | (d & Hd1 & Hd2) ]
                   | s1 z1 t1 t2 H2 [ IH2 | (d & Hd1 & Hd2) ]
                   ]; simpl in * |- *; try (destruct H1; fail).

    left; constructor 1; auto.
    red in H1; simpl in H1; destruct H1 as [ H1 H2 ].
    destruct H2 as [ H2 | H2 ].
    left; constructor 2; auto.
    right; exists (abt_leaf (inr (y1,s1))); simpl; split; auto; constructor 1; auto.
    
    left; constructor 3; auto.
    right; exists d2; split; auto; constructor 4; auto.
    right; exists d1; split; auto; constructor 3; auto.
    right; exists d1; split; auto; constructor 3; auto.
    
    left; constructor 5; auto.
    right; exists d; auto.
    
    left; constructor 6; auto.
    right; exists d; auto.
  Qed.

  Let good_or_disap l : 
                   good (embed_abtree SX ER RZ) l
                -> good (embed_abtree RX RY RZ) (map ev l)
                \/ exists x, In x l /\ has_disap x.
  Proof.
    induction 1 as [ l x1 x2 H1 H2 
                   | l x1 x2 [ H1 | (x & H1 & H2) ] ].
                   
    apply embed_prop in H2.
    destruct H2 as [ H2 | H2 ].
    left; simpl; constructor 1 with (ev x2); auto.
    rewrite in_map_iff; exists x2; auto.
    right; exists x2; simpl; tauto.
    
    left; simpl; constructor 2; auto.
    
    right; exists x; simpl; tauto.
  Qed.

  Notation R' := (embed_abtree SX ER RZ).

  Hypothesis HST : af R'.
  Hypothesis HST_t : af_t R'.
  
  Let HST' : bar (good R') nil.
  Proof.
    revert HST; apply af_eq_bar_good.
  Qed.
  
  Let HST_t' : bar_t (good R') nil.
  Proof.
    revert HST_t; apply af_t_bar_t.
  Qed.

  Notation U := (fun ws => Forall (good R') (list_fan (map ev_inv ws))).
  
  Let HU : bar U nil.
  Proof.
    apply bar_map_inv with (f := ev_inv) 
                           (Q := fun ll => Forall (good R') (list_fan ll)); auto.
    simpl; apply fan_on_list; auto.
    constructor; auto.
  Qed.
  
  Let HU_t : bar_t U nil.
  Proof.
    apply bar_t_map_inv with (f := ev_inv) 
                           (Q := fun ll => Forall (good R') (list_fan ll)); auto.
    simpl; apply fan_t_on_list; auto.
    constructor; auto.
  Qed.
  
  Let U_to_good ws : U ws -> good (embed_abtree RX RY RZ) (ws ++ (abt_cons m t) :: nil).
  Proof.
    set (Q l := good (embed_abtree RX RY RZ) (map ev l)).
    set (B a := has_disap a).
    
    intros H0.
    destruct (@hig_combi_principle _ (map ev_inv ws) Q B) as [ (p & Q1 & Q2) | (p & Q1 & Q2) ].
    
    (* condition of hig_combi_principle *)
    
    intros l Hl.
    apply list_fan_spec in Hl.
    simpl in H0; rewrite Forall_forall in H0.
    apply H0 in Hl; revert Hl.
    apply good_or_disap; auto.
    
    (* there is a good sequence p that evaluates to ws *)
      
    unfold Q in Q2.
    apply good_app_right.
    replace_with Q2.
    symmetry; apply Forall2_eq.
    revert Q1.
    rewrite Forall2_map_right, Forall2_map_left.
    apply Forall2_impl.
    intros ? ?; rewrite ev_inv_spec; auto.
    
    (* there is a fan in which every element contains a dispointing subtree *)
    
    unfold B in Q2.
    rewrite in_map_iff in Q1.
    destruct Q1 as (x & ? & Q1); subst p.
    apply in_split in Q1.
    destruct Q1 as (l1 & l2 & Q1).
    subst ws.
    rewrite <- app_assoc; simpl.
    apply good_app_left.
    constructor 1 with (abt_cons m t).
    apply in_or_app; right; left; auto.
    apply disap_prop.
    intros ?; rewrite <- ev_inv_spec; auto.
  Qed.

  Let bar_U_bar_good ws : bar U ws -> bar (good (embed_abtree RX RY RZ rlift abt_cons m t)) ws.
  Proof.
    induction 1.
    apply bar_lift_rel, in_bar_0; auto.
    apply in_bar_1; auto.
  Qed.  
  
  Let bar_t_U_bar_t_good ws : bar_t U ws -> bar_t (good (embed_abtree RX RY RZ rlift abt_cons m t)) ws.
  Proof.
    induction 1 using bar_t_rect.
    apply bar_t_lift_rel_rev, in_bar_t0; auto.
    apply in_bar_t1; auto.
  Qed.  
 
  Theorem abt_ev2_af : af (embed_abtree RX RY RZ rlift abt_cons m t).
  Proof.
    apply af_eq_bar_good; auto.
  Qed. 

  Theorem abt_ev2_af_t : af_t (embed_abtree RX RY RZ rlift abt_cons m t).
  Proof.
    apply bar_t_af_t; auto.
  Qed. 

End abtree_ev2.

Section abtree_ev3.

  Variables X Y Z : Type.

  Let Y' := (Y + (Z * abtree X Y Z + Z * abtree X Y Z))%type.
  
  Let ev := fix loop (t : abtree X Y' Z) : abtree X Y Z :=
    match t with
      | abt_leaf x                      => abt_leaf x 
      | abt_cons (inl y) t1             => abt_cons y (loop t1)
      | abt_cons (inr (inl (z,t1))) t2  => abt_pair z t1 (loop t2)
      | abt_cons (inr (inr (z,t2))) t1  => abt_pair z (loop t1) t2
      | abt_pair z t1 t2                => abt_pair z (loop t1) (loop t2)
    end.

  Let ev_inv := fix loop (t : abtree X Y Z) : list (abtree X Y' Z) :=
    match t with
      | abt_leaf x          => abt_leaf x::nil
      | abt_cons y t1       => map (abt_cons (inl y)) (loop t1)
      | abt_pair z t1 t2    => map (abt_cons (inr (inl (z,t1)))) (loop t2)
                            ++ map (abt_cons (inr (inr (z,t2)))) (loop t1)
                            ++ list_prod (abt_pair z) (loop t1) (loop t2) 
    end.

  Let ev_inv_spec s t : In t (ev_inv s) <-> s = ev t.
  Proof.
    revert t; induction s as [ x | y s1 Hs1 | z s1 Hs1 s2 Hs2 ]; intros t; split; simpl.

    intros [ ? | [] ]; subst; simpl; f_equal; auto.
    destruct t as [ | [|[(?,?)|(?,?)]] | ]; simpl; try discriminate 1.
    injection 1; intros; subst; tauto.
    
    intros H; apply in_map_iff in H.
    destruct H as (s & ? & H); subst; simpl; f_equal; apply Hs1; auto.
    destruct t as [ | [|[(?,?)|(?,?)]] | ]; simpl; try discriminate 1.
    intros H; injection H; clear H; intros; subst.
    apply in_map_iff.
    exists t; split; auto; apply Hs1; auto.
    
    intros H.
    apply in_app_or in H.
    destruct H as [ H | H ].

    apply in_map_iff in H.
    destruct H as (t1 & ? & Ht1); subst; simpl.
    f_equal; apply Hs2; auto.
    
    apply in_app_or in H.
    destruct H as [ H | H ].
    
    apply in_map_iff in H.
    destruct H as (t2 & ? & Ht2); subst; simpl.
    f_equal; apply Hs1; auto.
    
    apply list_prod_spec in H.
    destruct H as (t1 & t2 & Ht1 & Ht2 & ?); subst; simpl; f_equal; auto.
    apply Hs1; auto.
    apply Hs2; auto.
    
    destruct t as [ | [|[(?,?)|(?,?)]] | ]; simpl; try discriminate 1;
    intros H; injection H; clear H; intros; subst.

    apply in_or_app; left.
    apply in_map_iff.
    exists t; split; auto; apply Hs2; auto.
    
    apply in_or_app; right.
    apply in_or_app; left.
    apply in_map_iff.
    exists t; split; auto; apply Hs1; auto.
    
    apply in_or_app; right.
    apply in_or_app; right.
    apply list_prod_spec.
    exists t1, t2. repeat split.
    apply Hs1; auto.
    apply Hs2; auto.
  Qed.
  
  Notation analysis := (fun s a => s = ev a).
  
  Let analysis_finite s : finite (analysis s).
  Proof.
    exists (ev_inv s).
    intros; apply ev_inv_spec.
  Qed.
  
  Variables (RX : X -> X -> Prop) (RY : Y -> Y -> Prop) (RZ : Z -> Z -> Prop).

  Variable (m : Z) (t1 t2 : abtree X Y Z).

  Definition abt_rel3 :=
       sum_both RY (sum_both (rel_prod RZ (embed_abtree RX RY RZ rlift t1))
                             (rel_prod RZ (embed_abtree RX RY RZ rlift t2)) 
                   ).

  Notation SY := abt_rel3.
  Notation SZ := (RZ rlift m).
  
  Fact abt_rel3_af : af RY 
                  -> af RZ 
                  -> af (embed_abtree RX RY RZ rlift t1) 
                  -> af (embed_abtree RX RY RZ rlift t2) 
                  -> af abt_rel3.
  Proof.
    intros.
    repeat (apply af_sum_both; auto); apply af_product; auto.
  Qed. 

  Fact abt_rel3_af_t : af_t RY 
                    -> af_t RZ 
                    -> af_t (embed_abtree RX RY RZ rlift t1) 
                    -> af_t (embed_abtree RX RY RZ rlift t2) 
                    -> af_t abt_rel3.
  Proof.
    intros.
    repeat (apply af_t_sum_both; auto); apply af_t_prod; auto.
  Qed. 

  Let disapointing (s : abtree X Y' Z) := 
    match s with
      | abt_pair z _ _               => RZ m z
      | abt_cons (inl _) _           => False
      | abt_cons (inr (inl (_,v))) _ => embed_abtree RX RY RZ t1 v
      | abt_cons (inr (inr (_,v))) _ => embed_abtree RX RY RZ t2 v
      | abt_leaf _                   => False
    end.

  (* if every analysis of s contains a disappointing subtree then abt_cons m t embeds into s 
     this one is a bit difficult because it uses the combinatorial principle 

  *)

  Let has_disap s := exists a, disapointing a /\ sub_abtree a s.

  Let disap_prop s : analysis s inc1 has_disap
                  -> embed_abtree RX RY RZ (abt_pair m t1 t2) s.
  Proof.
    induction s as [ x | y s IHs | z s1 IHs1 s2 IHs2 ]; simpl; intros H.

    destruct (H (abt_leaf x)) as (a & H1 & H2).
    auto.
    inversion H2; subst.
    destruct H1.
    
    (* we decide on the analysis of s *) 
 
    destruct (@finite_decide _ (analysis s)) 
      with (R := fun a => has_disap a)
           (S := fun a => disapointing (abt_cons (inl y) a))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.

    intros x Hx.
    destruct (H (abt_cons (inl y) x)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_cons_inv in Hb2; 
      destruct Hb2 as [ | Hb2 ].
    subst b; right; auto.
    left; exists b; auto.
    
    constructor 4; apply IHs; auto.
    red in Hb2; destruct Hb2.
    
    (* we decide on the analysis of s1 *) 
 
    destruct (@finite_decide _ (analysis s1)) 
      with (R := fun a => has_disap a)
           (S := fun a => disapointing (abt_cons (inr (inr (z,s2))) a))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.

    intros x Hx.
    destruct (H (abt_cons (inr (inr (z,s2))) x)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_cons_inv in Hb2; 
      destruct Hb2 as [ | Hb2 ].
    subst b; right; auto.
    left; exists b; auto.
    
    constructor 5; apply IHs1; auto.
    red in Hb2; clear b Hb1.
    rename Hb2 into Hs2.
    
    (* we decide on the analysis of s2 *) 
 
    destruct (@finite_decide _ (analysis s2)) 
      with (R := fun a => has_disap a)
           (S := fun a => disapointing (abt_cons (inr (inl (z,s1))) a))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.

    intros x Hx.
    destruct (H (abt_cons (inr (inl (z,s1))) x)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_cons_inv in Hb2; 
      destruct Hb2 as [ | Hb2 ].
    subst b; right; auto.
    left; exists b; auto.
    
    constructor 6; apply IHs2; auto.
    red in Hb2; clear b Hb1.
    rename Hb2 into Hs1.

    (* we decide on the product of analysis of both s1 and s2 *)
    
    destruct (@finite_decide _ (pred_prod (analysis s1) (analysis s2)))
      with (R := fun c => has_disap (fst c) \/ has_disap (snd c))
           (S := fun c => disapointing (abt_pair z (fst c) (snd c)))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.
    apply finite_product; auto.
    
    intros (u1,u2) (H1 & H2); simpl in H1, H2.
    destruct (H (abt_pair z u1 u2)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_pair_inv in Hb2. 
      destruct Hb2 as [ | [ Hb2 | Hb2 ] ].
    subst b; right; auto.
    left; left; exists b; auto.
    left; right; exists b; auto.
    
    apply product_analysis in H0; auto.
    destruct H0 as [ H0 | H0 ].
    constructor 5; auto.
    constructor 6; auto.
    red in Hb2; clear b Hb1.
    
    constructor 3; auto.
  Qed.

  Let embed_prop s1 s2 : embed_abtree RX SY SZ s1 s2 
                      -> embed_abtree RX RY RZ (ev s1) (ev s2)
                      \/ has_disap s1.
  Proof.
    induction 1 as [ x1 x2 H1
                   | y1 s1 y2 s2 H1 H2 IH2
                   | z1 s1 u1 z2 s2 u2 H0 H1 IH1 H2 IH2 
                   | s1 y2 u2 H1 IH1
                   | s1 z2 u1 u2 H1 IH1
                   | s1 z2 u1 u2 H1 IH1
                   ]; simpl in * |- *.

    left; constructor 1; auto.
    
    destruct IH2 as [ IH2 | (d & H3 & H4) ].
    2: right; exists d; split; auto; constructor 2; auto.
    destruct y1 as [ y1 | [ (z1,u1) | (z1,u1) ] ];
    destruct y2 as [ y2 | [ (z2,u2) | (z2,u2) ] ]; try (destruct H1; fail).
    left; constructor 2; auto.
    destruct H1 as [ H3 [ H4 | H4 ] ]; simpl in H3, H4. 
    left; constructor 3; auto.
    right; exists (abt_cons (inr (inl (z1,u1))) s1); simpl; split; auto.
    constructor 1.
    destruct H1 as [ H3 [ H4 | H4 ] ]; simpl in H3, H4.
    left; constructor 3; auto.
    right; exists (abt_cons (inr (inr (z1,u1))) s1); simpl; split; auto.
    constructor 1.
    
    destruct IH1 as [ IH1 | (d & H3 & H4) ].
    2: right; exists d; split; auto; constructor 3; auto.
    destruct IH2 as [ IH2 | (d & H3 & H4) ].
    2: right; exists d; split; auto; constructor 4; auto.
    destruct H0.
    left; constructor 3; auto.
    right; exists (abt_pair z1 s1 u1); split; auto; constructor 1.
    
    destruct IH1 as [ IH1 | (d & H3 & H4) ].
    2: right; exists d; auto.
    destruct y2 as [ y2 | [ (z2,v2) | (z2,v2) ] ].
    left; constructor 4; auto.
    left; constructor 6; auto.
    left; constructor 5; auto.
    
    destruct IH1 as [ IH1 | (d & H3 & H4) ].
    2: right; exists d; auto.
    left; constructor 5; auto.
    
    destruct IH1 as [ IH1 | (d & H3 & H4) ].
    2: right; exists d; auto.
    left; constructor 6; auto.
  Qed.

  Let good_or_disap l : 
                   good (embed_abtree RX SY SZ) l
                -> good (embed_abtree RX RY RZ) (map ev l)
                \/ exists x, In x l /\ has_disap x.
  Proof.
    induction 1 as [ l x1 x2 H1 H2 
                   | l x1 x2 [ H1 | (x & H1 & H2) ] ].
                   
    apply embed_prop in H2.
    destruct H2 as [ H2 | H2 ].
    left; simpl; constructor 1 with (ev x2); auto.
    rewrite in_map_iff; exists x2; auto.
    right; exists x2; simpl; tauto.
    
    left; simpl; constructor 2; auto.
    
    right; exists x; simpl; tauto.
  Qed.

  Notation R' := (embed_abtree RX SY SZ).

  Hypothesis HST : af R'.
  Hypothesis HST_t : af_t R'.
  
  Let HST' : bar (good R') nil.
  Proof.
    revert HST; apply af_eq_bar_good.
  Qed.

  Let HST_t' : bar_t (good R') nil.
  Proof.
    revert HST_t; apply af_t_bar_t.
  Qed.

  Notation U := (fun ws => Forall (good R') (list_fan (map ev_inv ws))).
  
  Let HU : bar U nil.
  Proof.
    apply bar_map_inv with (f := ev_inv) 
                           (Q := fun ll => Forall (good R') (list_fan ll)); auto.
    simpl; apply fan_on_list; auto.
    constructor; auto.
  Qed.

  Let HU_t : bar_t U nil.
  Proof.
    apply bar_t_map_inv with (f := ev_inv) 
                           (Q := fun ll => Forall (good R') (list_fan ll)); auto.
    simpl; apply fan_t_on_list; auto.
    constructor; auto.
  Qed.
  
  Let U_to_good ws : U ws -> good (embed_abtree RX RY RZ) (ws ++ (abt_pair m t1 t2) :: nil).
  Proof.
    set (Q l := good (embed_abtree RX RY RZ) (map ev l)).
    set (B a := has_disap a).
    
    intros H0.
    destruct (@hig_combi_principle _ (map ev_inv ws) Q B) as [ (p & Q1 & Q2) | (p & Q1 & Q2) ].
    
    (* condition of hig_combi_principle *)
    
    intros l Hl.
    apply list_fan_spec in Hl.
    simpl in H0; rewrite Forall_forall in H0.
    apply H0 in Hl; revert Hl.
    apply good_or_disap; auto.
    
    (* there is a good sequence p that evaluates to ws *)
      
    unfold Q in Q2.
    apply good_app_right.
    replace_with Q2.
    symmetry; apply Forall2_eq.
    revert Q1.
    rewrite Forall2_map_right, Forall2_map_left.
    apply Forall2_impl.
    intros ? ?; rewrite ev_inv_spec; auto.
    
    (* there is a fan in which every element contains a dispointing subtree *)
    
    unfold B in Q2.
    rewrite in_map_iff in Q1.
    destruct Q1 as (x & ? & Q1); subst p.
    apply in_split in Q1.
    destruct Q1 as (l1 & l2 & Q1).
    subst ws.
    rewrite <- app_assoc; simpl.
    apply good_app_left.
    constructor 1 with (abt_pair m t1 t2).
    apply in_or_app; right; left; auto.
    apply disap_prop.
    intros ?; rewrite <- ev_inv_spec; auto.
  Qed.
  
  Let bar_U_bar_good ws : bar U ws -> bar (good (embed_abtree RX RY RZ rlift abt_pair m t1 t2)) ws.
  Proof.
    induction 1.
    apply bar_lift_rel, in_bar_0; auto.
    apply in_bar_1; auto.
  Qed.  

  Let bar_t_U_bar_t_good ws : bar_t U ws -> bar_t (good (embed_abtree RX RY RZ rlift abt_pair m t1 t2)) ws.
  Proof.
    induction 1 using bar_t_rect.
    apply bar_t_lift_rel_rev, in_bar_t0; auto.
    apply in_bar_t1; auto.
  Qed. 
   
  Theorem abt_ev3_af : af (embed_abtree RX RY RZ rlift abt_pair m t1 t2).
  Proof.
    apply af_eq_bar_good; auto.
  Qed. 

  Theorem abt_ev3_af_t : af_t (embed_abtree RX RY RZ rlift abt_pair m t1 t2).
  Proof.
    apply bar_t_af_t; auto.
  Qed. 

End abtree_ev3.

Section abtree_ev4.

  Variables X Y Z : Type.

  Let Y' := (Y + (Z * abtree X Y Z + Z * abtree X Y Z))%type.
  
  Let ev := fix loop (t : abtree X Y' ES) : abtree X Y Z :=
    match t with
      | abt_leaf x                      => abt_leaf x 
      | abt_cons (inl y) t1             => abt_cons y (loop t1)
      | abt_cons (inr (inl (z,t1))) t2  => abt_pair z t1 (loop t2)
      | abt_cons (inr (inr (z,t2))) t1  => abt_pair z (loop t1) t2
      | abt_pair z _ _                  => Empty_set_rect _ z
    end.

  Let ev_inv := fix loop (t : abtree X Y Z) : list (abtree X Y' ES) :=
    match t with
      | abt_leaf x          => abt_leaf x::nil
      | abt_cons y t1       => map (abt_cons (inl y)) (loop t1)
      | abt_pair z t1 t2    => map (abt_cons (inr (inl (z,t1)))) (loop t2)
                            ++ map (abt_cons (inr (inr (z,t2)))) (loop t1)
    end.

  Let ev_inv_spec s t : In t (ev_inv s) <-> s = ev t.
  Proof.
    revert t; induction s as [ x | y s1 Hs1 | z s1 Hs1 s2 Hs2 ]; intros t; split; simpl.

    intros [ ? | [] ]; subst; simpl; f_equal; auto.
    destruct t as [ | [|[(?,?)|(?,?)]] | [] ]; simpl; try discriminate 1.
    injection 1; intros; subst; tauto.
    
    intros H; apply in_map_iff in H.
    destruct H as (s & ? & H); subst; simpl; f_equal; apply Hs1; auto.
    destruct t as [ | [|[(?,?)|(?,?)]] | [] ]; simpl; try discriminate 1.
    intros H; injection H; clear H; intros; subst.
    apply in_map_iff.
    exists t; split; auto; apply Hs1; auto.
    
    intros H.
    apply in_app_or in H.
    destruct H as [ H | H ].

    apply in_map_iff in H.
    destruct H as (t1 & ? & Ht1); subst; simpl.
    f_equal; apply Hs2; auto.
   
    apply in_map_iff in H.
    destruct H as (t2 & ? & Ht2); subst; simpl.
    f_equal; apply Hs1; auto.
    
    destruct t as [ | [|[(?,?)|(?,?)]] | [] ]; simpl; try discriminate 1;
    intros H; injection H; clear H; intros; subst.

    apply in_or_app; left.
    apply in_map_iff.
    exists t; split; auto; apply Hs2; auto.
    
    apply in_or_app; right.
    apply in_map_iff.
    exists t; split; auto; apply Hs1; auto.
  Qed.
  
  Notation analysis := (fun s a => s = ev a).
  
  Let analysis_finite s : finite (analysis s).
  Proof.
    exists (ev_inv s).
    intros; apply ev_inv_spec.
  Qed.
 
  Variables (RX : X -> X -> Prop) (RY : Y -> Y -> Prop) (RZ : Z -> Z -> Prop).

  Variable (m : Z) (t1 t2 : abtree X Y Z).

  Definition abt_rel4 :=
       sum_both RY (sum_both (rel_prod RZ (embed_abtree RX RY RZ rlift t1))
                             (rel_prod RZ (embed_abtree RX RY RZ rlift t2)) 
                   ).

  Notation SY := abt_rel4.
  
  Fact abt_rel4_af : af RY 
                  -> full RZ 
                  -> af (embed_abtree RX RY RZ rlift t1) 
                  -> af (embed_abtree RX RY RZ rlift t2) 
                  -> af abt_rel4.
  Proof.
    intros.
    assert (af RZ). 
      constructor 1; auto.
    repeat (apply af_sum_both; auto); apply af_product; auto.
  Qed. 

  Fact abt_rel4_af_t : af_t RY 
                  -> full RZ 
                  -> af_t (embed_abtree RX RY RZ rlift t1) 
                  -> af_t (embed_abtree RX RY RZ rlift t2) 
                  -> af_t abt_rel4.
  Proof.
    intros.
    assert (af_t RZ). 
      apply in_af_t0; auto.
    repeat (apply af_t_sum_both; auto); apply af_t_prod; auto.
  Qed. 

  Let disapointing (s : abtree X Y' ES) := 
    match s with
      | abt_pair z _ _               => False
      | abt_cons (inl _) _           => False
      | abt_cons (inr (inl (_,v))) _ => embed_abtree RX RY RZ t1 v
      | abt_cons (inr (inr (_,v))) _ => embed_abtree RX RY RZ t2 v
      | abt_leaf _                   => False
    end.

  Hypothesis HRZ : full RZ.

  Let has_disap s := exists a, disapointing a /\ sub_abtree a s.

  Let disap_prop s : analysis s inc1 has_disap
                  -> embed_abtree RX RY RZ (abt_pair m t1 t2) s.
  Proof.
    induction s as [ x | y s IHs | z s1 IHs1 s2 IHs2 ]; simpl; intros H.

    destruct (H (abt_leaf x)) as (a & H1 & H2).
    auto.
    inversion H2; subst.
    destruct H1.
    
    (* we decide on the analysis of s *) 
 
    destruct (@finite_decide _ (analysis s)) 
      with (R := fun a => has_disap a)
           (S := fun a => disapointing (abt_cons (inl y) a))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.

    intros x Hx.
    destruct (H (abt_cons (inl y) x)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_cons_inv in Hb2; 
      destruct Hb2 as [ | Hb2 ].
    subst b; right; auto.
    left; exists b; auto.
    
    constructor 4; apply IHs; auto.
    red in Hb2; destruct Hb2.
    
    (* we decide on the analysis of s1 *) 
 
    destruct (@finite_decide _ (analysis s1)) 
      with (R := fun a => has_disap a)
           (S := fun a => disapointing (abt_cons (inr (inr (z,s2))) a))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.

    intros x Hx.
    destruct (H (abt_cons (inr (inr (z,s2))) x)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_cons_inv in Hb2; 
      destruct Hb2 as [ | Hb2 ].
    subst b; right; auto.
    left; exists b; auto.
    
    constructor 5; apply IHs1; auto.
    red in Hb2; clear b Hb1.
    rename Hb2 into Hs2.
    
    (* we decide on the analysis of s2 *) 
 
    destruct (@finite_decide _ (analysis s2)) 
      with (R := fun a => has_disap a)
           (S := fun a => disapointing (abt_cons (inr (inl (z,s1))) a))
      as [ H0 | (b & Hb1 & Hb2) ]; auto.

    intros x Hx.
    destruct (H (abt_cons (inr (inl (z,s1))) x)) as (b & Hb1 & Hb2).
    subst; auto.
    apply sub_abtree_cons_inv in Hb2; 
      destruct Hb2 as [ | Hb2 ].
    subst b; right; auto.
    left; exists b; auto.
    
    constructor 6; apply IHs2; auto.
    red in Hb2; clear b Hb1.
    rename Hb2 into Hs1.
    
    constructor 3; auto.
  Qed.
  
  Let embed_prop s1 s2 : embed_abtree RX SY ER s1 s2 
                      -> embed_abtree RX RY RZ (ev s1) (ev s2)
                      \/ has_disap s1.
  Proof.
    induction 1 as [ x1 x2 H1
                   | y1 s1 y2 s2 H1 H2 IH2
                   | [] s1 u1 z2 s2 u2 H0 H1 IH1 H2 IH2 
                   | s1 y2 u2 H1 IH1
                   | s1 [] u1 u2 H1 IH1
                   | s1 [] u1 u2 H1 IH1
                   ]; simpl in * |- *.

    left; constructor 1; auto.
    
    destruct IH2 as [ IH2 | (d & H3 & H4) ].
    2: right; exists d; split; auto; constructor 2; auto.
    destruct y1 as [ y1 | [ (z1,u1) | (z1,u1) ] ];
    destruct y2 as [ y2 | [ (z2,u2) | (z2,u2) ] ]; try (destruct H1; fail).
    left; constructor 2; auto.
    destruct H1 as [ H3 [ H4 | H4 ] ]; simpl in H3, H4. 
    left; constructor 3; auto.
    right; exists (abt_cons (inr (inl (z1,u1))) s1); simpl; split; auto.
    constructor 1.
    destruct H1 as [ H3 [ H4 | H4 ] ]; simpl in H3, H4.
    left; constructor 3; auto.
    right; exists (abt_cons (inr (inr (z1,u1))) s1); simpl; split; auto.
    constructor 1.
    
    destruct IH1 as [ IH1 | (d & H3 & H4) ].
    2: right; exists d; auto.
    destruct y2 as [ y2 | [ (z2,v2) | (z2,v2) ] ].
    left; constructor 4; auto.
    left; constructor 6; auto.
    left; constructor 5; auto.
  Qed.

  Let good_or_disap l : 
                   good (embed_abtree RX SY ER) l
                -> good (embed_abtree RX RY RZ) (map ev l)
                \/ exists x, In x l /\ has_disap x.
  Proof.
    induction 1 as [ l x1 x2 H1 H2 
                   | l x1 x2 [ H1 | (x & H1 & H2) ] ].
                   
    apply embed_prop in H2.
    destruct H2 as [ H2 | H2 ].
    left; simpl; constructor 1 with (ev x2); auto.
    rewrite in_map_iff; exists x2; auto.
    right; exists x2; simpl; tauto.
    
    left; simpl; constructor 2; auto.
    
    right; exists x; simpl; tauto.
  Qed.

  Notation R' := (embed_abtree RX SY ER).

  Hypothesis HST : af R'.
  Hypothesis HST_t : af_t R'.

  Let HST' : bar (good R') nil.
  Proof.
    revert HST; apply af_eq_bar_good.
  Qed.

  Let HST_t' : bar_t (good R') nil.
  Proof.
    revert HST_t; apply af_t_bar_t.
  Qed.

  Notation U := (fun ws => Forall (good R') (list_fan (map ev_inv ws))).
  
  Let HU : bar U nil.
  Proof.
    apply bar_map_inv with (f := ev_inv) 
                           (Q := fun ll => Forall (good R') (list_fan ll)); auto.
    simpl; apply fan_on_list; auto.
    constructor; auto.
  Qed.

  Let HU_t : bar_t U nil.
  Proof.
    apply bar_t_map_inv with (f := ev_inv) 
                           (Q := fun ll => Forall (good R') (list_fan ll)); auto.
    simpl; apply fan_t_on_list; auto.
    constructor; auto.
  Qed.
  
  Let U_to_good ws : U ws -> good (embed_abtree RX RY RZ) (ws ++ (abt_pair m t1 t2) :: nil).
  Proof.
    set (Q l := good (embed_abtree RX RY RZ) (map ev l)).
    set (B a := has_disap a).
    
    intros H0.
    destruct (@hig_combi_principle _ (map ev_inv ws) Q B) as [ (p & Q1 & Q2) | (p & Q1 & Q2) ].
    
    (* condition of hig_combi_principle *)
    
    intros l Hl.
    apply list_fan_spec in Hl.
    simpl in H0; rewrite Forall_forall in H0.
    apply H0 in Hl; revert Hl.
    apply good_or_disap; auto.
    
    (* there is a good sequence p that evaluates to ws *)
      
    unfold Q in Q2.
    apply good_app_right.
    replace_with Q2.
    symmetry; apply Forall2_eq.
    revert Q1.
    rewrite Forall2_map_right, Forall2_map_left.
    apply Forall2_impl.
    intros ? ?; rewrite ev_inv_spec; auto.
    
    (* there is a fan in which every element contains a dispointing subtree *)
    
    unfold B in Q2.
    rewrite in_map_iff in Q1.
    destruct Q1 as (x & ? & Q1); subst p.
    apply in_split in Q1.
    destruct Q1 as (l1 & l2 & Q1).
    subst ws.
    rewrite <- app_assoc; simpl.
    apply good_app_left.
    constructor 1 with (abt_pair m t1 t2).
    apply in_or_app; right; left; auto.
    apply disap_prop.
    intros ?; rewrite <- ev_inv_spec; auto.
  Qed.
  
  Let bar_U_bar_good ws : bar U ws -> bar (good (embed_abtree RX RY RZ rlift abt_pair m t1 t2)) ws.
  Proof.
    induction 1.
    apply bar_lift_rel, in_bar_0; auto.
    apply in_bar_1; auto.
  Qed.  

  Let bar_t_U_bar_t_good ws : bar_t U ws -> bar_t (good (embed_abtree RX RY RZ rlift abt_pair m t1 t2)) ws.
  Proof.
    induction 1 using bar_t_rect.
    apply bar_t_lift_rel_rev, in_bar_t0; auto.
    apply in_bar_t1; auto.
  Qed. 
   
  Theorem abt_ev4_af : af (embed_abtree RX RY RZ rlift abt_pair m t1 t2).
  Proof.
    apply af_eq_bar_good; auto.
  Qed. 

  Theorem abt_ev4_af_t : af_t (embed_abtree RX RY RZ rlift abt_pair m t1 t2).
  Proof.
    apply bar_t_af_t; auto.
  Qed. 

End abtree_ev4.
