(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.
Require Import ChoiceFacts.

Require Omega.

Require Import rel_utils.
Require Import list_utils.

Require Import good_base.

Require Import wft.
Require Import bar.
Require Import af.
Require Import af_bar.
Require Import af_t.

Section af_to_af_t.

  Variable (X : Type).
  
  Implicit Type (R : X -> X -> Prop).

  Hypothesis FC : FunctionalChoice_on X (wft X).

  Fact af_to_af_securedby R : af R -> exists t, af_securedby R t.
  Proof.
    induction 1 as [ R HR | R HR IHR ].
    
    exists in_wft_zt; auto.
    apply FC in IHR.
    destruct IHR as (f & Hf).
    exists (in_wft_sup f); simpl; auto.
  Qed.

  Fact af_to_af_t R : af R <-> inhabited (af_t R).
  Proof.
    split.

    intros H.
    apply af_to_af_securedby in H.
    destruct H as (t & Ht).
    exists; exists t; auto.
    
    intros [ ].
    induction 1 using af_t_rect.
    constructor 1; auto.
    constructor 2; auto.
  Qed.

End af_to_af_t.

Require Import Classical.

Require Import list_dep_choice.

Section af_classic.

  Variable (X : Type) (R : X -> X -> Prop).

  Fact af_classic : (forall f, exists i j, i < j /\ R (f i) (f j)) -> af R.
  Proof.
    intros H; apply NNPP.
    contradict H; intros Hf.
    rewrite af_eq_bar_good in H.
    destruct (functional_list_dep_choice _ (fun l => ~ bar (good R) l)) as (f & Hg); auto.
    intros; apply classic.
    intros l Hl.
    apply not_all_ex_not.
    contradict Hl.
    constructor 2; auto.
    destruct (Hf f) as (i & j & H1 & H2).
    apply (Hg (S j)).
    simpl.
    constructor 1.
    constructor 1 with (f i); auto.
    rewrite pfx_rev_In.
    exists i; auto.
  Qed.

  Fact af_classic_inv : ~ af R -> exists f, forall i j, i < j -> ~ R (f i) (f j).
  Proof.
    intros H.
    apply NNPP.
    contradict H.
    apply af_classic.
    intros f.
    generalize (not_ex_all_not _ _ H f); clear H; intros H.
    apply not_all_ex_not in H; destruct H as (i & H).
    apply not_all_ex_not in H; destruct H as (j & H).
    apply imply_to_and in H.
    destruct H as [ H1 H2 ].
    exists i, j; split; auto.
    apply NNPP; auto.
  Qed.

End af_classic.

Section application_to_Ramsey.

  Variables (X : Type) (R : X -> X -> Prop).

  Let R0 n m := n = m \/ ~ R n m.
  Let R1 n m := n = m \/   R n m.

  Fact R01 n m : (R0 n m /\ R1 n m) <-> n = m.
  Proof.
    unfold R0, R1.
    split.
    intros [ [ | ] [ | ] ]; subst; auto.
    intro; contradict H; auto.
    intros; subst; tauto.
  Qed.

  Definition inj (f : nat -> X) := forall i j, f i = f j -> i = j.

  Definition homo (f : nat -> X) := (forall n m, n < m ->   R (f n) (f m)) 
                                 \/ (forall n m, n < m -> ~ R (f n) (f m)). 

  (* this is on possible definition of infinite *) 

  Hypothesis X_infinite : exists f, inj f.

  (* this is another one, they are classically equiv. but not intuitionistically *)

  Let X_infty : ~ af (@eq X).
  Proof.
    intros H.
    destruct X_infinite as (f & Hf).
    destruct (af_inf_chain f H) as (i & j & H1 & H2).
    apply Hf in H2.
    omega.
  Qed.

  Fact not_af_R0 : ~ af R0 -> exists f, forall i j, i < j -> (f i <> f j /\ R (f i) (f j)).
  Proof.
    intros H.
    apply af_classic_inv in H.
    destruct H as (f & Hf).
    exists f.
    intros i j H.
    specialize (Hf _ _ H).
    unfold R0 in Hf.
    apply not_or_and in Hf.
    destruct Hf.
    split; auto.
    apply NNPP; auto.
  Qed.
  
  Fact not_af_R1 : ~ af R1 -> exists f, forall i j, i < j -> (f i <> f j /\ ~ R (f i) (f j)).
  Proof.
    intros H.
    apply af_classic_inv in H.
    destruct H as (f & Hf).
    exists f.
    intros i j H.
    specialize (Hf _ _ H).
    unfold R0 in Hf.
    apply not_or_and in Hf; auto.
  Qed.
   
  Theorem Ramsey : exists f, inj f /\ homo f.
  Proof.
    (* classically, either R0 or R1 is not almost full, otherwise R0 /\ R1 is *)
  
    assert (~ af R0 \/ ~ af R1) as H.
      apply not_and_or.
      contradict X_infty.
      apply af_inc with (R := fun n m => R0 n m /\ R1 n m).
      intros ? ?; apply R01.
      apply af_intersection; apply X_infty.
    
    destruct H as [ H | H ].

    (* if R0 is not almost full that we build an infinite R-homogeneous subset of X *)

    apply not_af_R0 in H.
    destruct H as (f & Hf).
    exists f; split.
    intros i j H.
    destruct (lt_eq_lt_dec i j) as [ [ G |  ] | G ]; auto.
    apply Hf, proj1 in G; contradict G; auto.
    apply Hf, proj1 in G; contradict G; auto.
    left; intros ? ? ?; apply Hf; auto.

    (* if R1 is not almost full that we build an infinite (not R)-homogeneous subset of X *)
    
    apply not_af_R1 in H.
    destruct H as (f & Hf).
    exists f; split.
    intros i j H.
    destruct (lt_eq_lt_dec i j) as [ [ G |  ] | G ]; auto.
    apply Hf, proj1 in G; contradict G; auto.
    apply Hf, proj1 in G; contradict G; auto.
    right; intros ? ? ?; apply Hf; auto.
  Qed.
    
End application_to_Ramsey.
