(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Min.
Require Import List.
Require Import Relations.
Require Import Wf.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import subword.

Require Import wft.
Require Import af_t.
Require Import afs_t.

Require Import tree.
Require Import tree_embeddings.

Require Import tree_kruskal_fix.
Require Import tree_kruskal_ev1.
Require Import tree_kruskal_ev2.
Require Import tree_kruskal_ev3.
Require Import tree_kruskal_ev4.

Require Import higman_ktree_af_t.

Set Implicit Arguments.

Section kruskal_ktree_rec.

  Variable X : Type.
  Notation T := (ktree_fix X).

  Theorem kruskal_ktree_rec (s : nt_stump T) :
              forall k, k = nts_char s
           -> forall (P : nat -> T -> Prop),
                     (forall n, ~ P n in_ktf_u)
                  -> (forall n x, { P n x } + { ~ P n x }) 
                  -> (forall n, k <= n -> P k = P n)
           -> forall (R : nat -> relation T),
                     (forall n, n <= k -> afs_owft_sec (nts_seq s n) (P n) (R n))
                  ->  afs_t (wfptree P) (emb_tree_upto k R).
  Proof.
(* the outermost induction is over non trivial stumps with the more facile
   well founded relation *)

    induction s as [ s IHs ] using nt_stump_rect.
    intros k Hk P HP1 HP2 HP3 R HR1.
    assert (forall t, { wfptree P t } + { ~ wfptree P t }) as HP4.
      apply wfptree_dec, HP2.
    
(* now starts the nested induction over the well-formed tree *)

    apply in_afs_t1; auto.
    induction 1 as [ m lt H_m_lt IHt ] using wfptree_rect.
    
    case_eq (length lt).

(* case where lt is empty, ie we have a leaf *)
    
    intros Hl_lt.
    destruct lt.
    2: discriminate Hl_lt.
    generalize H_m_lt.
    intros Hm0.
    rewrite wfptree_fix in H_m_lt.
    destruct H_m_lt as [ Hm1 _ ]; simpl in Hm1.
    generalize (HR1 0); intros HR0.
    spec_all HR0.
    omega.

    case_eq (nts_seq s 0); [ intros [ | f ] | ]; intros Hs0; 
      rewrite Hs0 in HR0; simpl in HR0.
    Focus 3.

  (* s 0 = None, then P 0 is empty *)
    
    contradict Hm1; apply HR0.

  (* s 0 = Some in_wft_zt, then R 0 is full on P 0 *)
    
    apply in_afs_t0.
    intros x y Hx Hy; right.
    clear y Hy.
    destruct (tree_leaf x) as (l & Hl).
    unfold wfptree in Hx.
    rewrite tree_fall_sub_tree in Hx.    
    generalize Hl.     
    apply emb_sub_tree_upto.
    apply in_emb_tree_upto_nil; auto.
    apply HR0; auto.
    apply Hx with (1 := Hl).

  (* s 0 = Some (in_wft_sup f), then R rlift x is secured by f x *)
    
    set (s' n := match n with 0 => Some (f m) | S n => nts_seq s (S n) end).
    assert (Hs' : s' k <> None /\ forall i, k < i -> s' i = None).
      unfold s'.
      split.
      destruct k.
      discriminate.
      rewrite Hk.
      apply nts_prop1.
      intros [ | i ].
      omega.
      rewrite Hk.
      apply nts_prop2.
    set (s1 := existT _ s' (exist _ k Hs') : nt_stump T).    
    set (R' n := match n with 0 => R 0 rlift m | S n => R (S n) end).
    specialize (IHs s1); spec_all IHs.
    
    (* s1 is more facile that s *)
    
    red; right.
    rewrite <- Hk.
    unfold s1; simpl; split; auto.
    exists 0; split.
    unfold s'; rewrite Hs0.
    simpl; apply in_wft_lt.
    intros [ | j ]; simpl; auto; omega.
    
    specialize (IHs k eq_refl P); spec_all IHs; auto.
    specialize (IHs R'); spec_all IHs; auto.
    
    (* s1 secures R' upto k *)
 
    intros [ | n ]; unfold s1; simpl.
    intros; auto.
    apply HR1.
    
    (* now we can use the induction hypothesis *)

    revert IHs; apply afs_t_incl; auto.
    intros t1 t2.
    induction 1 as [ t1 t2 x ll Hll H IH 
                   | x y ll mm Hll Ha Hb IH
                   | x y ll mm Hll Ha Hb IH 
                   ] using emb_tree_upto_ind.

    destruct IH as [ IH | IH ].
    left; apply in_emb_tree_upto_0 with t2; auto.
    right; auto.

    apply Forall2_decide_ind in IH.
    destruct IH as [ IH | IH ].

    destruct ll; simpl in Ha.
    destruct Ha as [ Ha | Ha ].
    inversion_clear IH.
    left; apply in_emb_tree_upto_nil; auto.
    right; apply in_emb_tree_upto_nil; auto.
    left; apply in_emb_tree_upto_1; auto.

    rewrite Exists2_exists_left in IH.
    destruct IH as (IH & t & G4 & G5).
    right; apply in_emb_tree_upto_0 with t; auto.
    
    apply subword_cup in IH.
    destruct IH as [ IH | IH ].
    destruct k; simpl in Ha.
    destruct Ha as [ Ha | Ha ].
    left; apply in_emb_tree_upto_2; auto.
    right; apply in_emb_tree_upto_2; auto; apply subword_nil_true.
    left; apply in_emb_tree_upto_2; auto.
    
    destruct IH as (u & _ & G1 & _ & G2).
    right; apply in_emb_tree_upto_0 with u; auto.
       
(* case where lt is NOT empty, ie NOT a leaf *)
    
    intros i Hi.
    generalize H_m_lt.
    intros Hm0.
    rewrite wfptree_fix in Hm0.
    destruct Hm0 as [ Hm1 Hm2 ].
    rewrite Hi in Hm1.    
    generalize (HR1 (S i)); intros HRSi.

    case (lt_eq_lt_dec k (S i)); [ intros [ H_k_i | H_k_i ] | intros H_k_i ].
        
(* case where S i = length lt and k <= i 
   in that case nts_seq s (S i) must be None 
*)
    
    apply le_S_n in H_k_i.
    clear HRSi.
    generalize (HR1 _ (le_refl _)); intros HRk.
    
    case_eq (nts_seq s k); [ intros [ | f ] | ]; intros Hs0;
      rewrite Hs0 in HRk; simpl in HRk.
    Focus 3.

  (* s k = None, is not possible because char s = k *)
        
    contradict Hs0; rewrite Hk; apply nts_prop1.
    
  (* s k = Some in_wft_zt, then R k is full on P k *)
    
    set (P' := @ktf_prop2 X k P i).
    assert (G0 : afs_t (P' i) (ktf_rel2_i k i lt R)).
      apply afs_t_ktf_rel2_i; auto.
      rewrite min_r; auto.
      apply in_afs_t0; auto.
      apply in_afs_t0; auto.
    set (R' := @ktf_rel2 _ k i lt R).
  
    generalize (@higman_wfptree _ (S i) P' R'); intros Hig; spec_all Hig.
    
    apply ktf_prop2_dec; auto.
    intros n ? Hn; unfold P'; rewrite ktf_prop2_gt; auto.
    intros n Hn.
    unfold P', R'.
    destruct (ktf_pos2 k i n) as [[[(?&?)|]|H]|]; try omega.

    rewrite ktf_prop2_lt; try omega.
    rewrite ktf_rel2_eq_0; auto.
    apply afs_owft_afs_t with (nts_seq s n), HR1; omega.
    
    rewrite ktf_prop2_lt; try omega.
    rewrite ktf_rel2_eq_1; auto.
    rewrite <- HP3; try omega.
    apply afs_owft_afs_t with (nts_seq s k), HR1; omega.
   
    rewrite H, ktf_rel2_eq_2; auto.

    (* now Hig is proof that emb_ntree R' is af on wfptree P' *)
    
    apply afs_t_impl_af_t in Hig.
    apply af_t_impl_afs_t; auto.
    revert Hig; apply ktf_ev2_af_t with (H_m_lt := H_m_lt); auto.

  (* s k = Some (in_wft_sup f), then R k rlift x is secured by f x *)
    
    set (P' := @ktf_prop1 X k P i).
    assert (G0 : afs_t (P' i) (ktf_rel1_i k i lt R)).
      apply afs_t_ktf_rel1_i; auto.
      generalize (HR1 _ (le_min_r i k)); apply afs_owft_afs_t.
      generalize (HR1 _ (le_refl _)); apply afs_owft_afs_t.
    apply afs_t_afs_owft in G0.
    destruct G0 as (owfti & Howfti).
    assert (Hki : k <= S i).
      omega.
    set (R' := @ktf_rel1 _ k i m lt R).
    set (s' j := match (ktf_pos1 k i j) with
                   | inleft (inleft (inleft (left _)))   => nts_seq s j
                   | inleft (inleft (inleft (right _)))  => nts_seq s k
                   | inleft (inleft (inright _))         => owfti 
                   | inleft (inright _)                  => Some (f m)
                   | inright _                           => None
                 end).
    assert (Hs' : s' (S i) <> None /\ forall j, S i < j -> s' j = None).
      unfold s'.
      split.
      destruct (ktf_pos1 k i (S i)) as [[[[|]|]|]|]; try omega.
      discriminate.
      intros j Hj.
      destruct (ktf_pos1 k i j) as [[[[|]|]|]|]; try omega; auto.
    set (s1 := existT _ s' (exist _ (S i) Hs') : nt_stump T).    
    specialize (IHs s1); spec_all IHs; auto.

    (* s' is more facile than s *)
    
    red; left. 
    rewrite <- Hk, Hs0.
    unfold s1; simpl.
    unfold s'.
    destruct (ktf_pos1 k i (S i)) as [[[[|]|]|]|]; try omega.
    red; apply in_wft_lt.
    
    specialize (IHs (S i) eq_refl P'); spec_all IHs; auto.
    
    apply ktf_prop1_u; auto.
    apply ktf_prop1_dec; auto.
    intros n Hn.
    unfold P'.
    do 2 (rewrite ktf_prop1_neq; try omega).
    rewrite <- (HP3 n); try omega.
    rewrite <- HP3; try omega.
    auto.
    
    specialize (IHs R'); spec_all IHs; auto.

    (* we show that s' secures P'/R' *)
    
    intros n Hn.
    simpl; unfold s', P', R'.
    destruct (ktf_pos1 k i n) as [[[[(?&?)|]|]|]|]; try omega.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_0; auto.
    apply HR1; omega.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_1; auto.
    rewrite <- HP3; try omega.
    apply HR1; omega.
    rewrite ktf_rel1_eq_2; auto.
    subst n; apply Howfti.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_3; auto.
    simpl; rewrite <- HP3; try omega.
    apply HRk.
    rewrite <- HP3 in Hm1; auto; omega.

    (* now IHs is proof that emb_tree_upto (S i) R' is af on wfptree P' *)
   
    apply afs_t_impl_af_t in IHs.
    apply af_t_impl_afs_t; auto.
    revert IHs; apply ktf_ev1_af_t with (H_m_lt := H_m_lt); auto.

(* case k = S i *)
    
    spec_all HRSi.
    omega.
    
    case_eq (nts_seq s (S i)); [ intros [ | f ] | ]; intros Hs0;
      rewrite Hs0 in HRSi; simpl in HRSi.
    Focus 3.

  (* s (S i) = None, is not possible because m belongs to P (S i) *)
    
    contradict Hm1; auto.
    
  (* s (S i) = Some in_wft_zt, then R (S i) is full on P (S i) *)

    rewrite <- H_k_i in HRSi.
    set (P' := @ktf_prop2 X k P i).
    assert (G0 : afs_t (P' i) (ktf_rel2_i k i lt R)).
      apply afs_t_ktf_rel2_i; auto.
      omega.
      rewrite min_l.
      specialize (HR1 i); spec_all HR1.
      omega.
      revert HR1; apply afs_owft_afs_t.
      omega.
      apply in_afs_t0; auto.
    set (R' := @ktf_rel2 _ k i lt R).

    (* we can use Higman theorem in that case *)
  
    generalize (@higman_wfptree _ (S i) P' R'); intros Hig; spec_all Hig.

    apply (@ktf_prop2_dec _ k); auto; omega.
    intros n ? Hn; unfold P'; rewrite (ktf_prop2_gt (k := k)); auto; omega.
    intros n Hn.
    unfold P', R'.
    destruct (ktf_pos2 k i n) as [[[(?&?)|]|H]|]; try omega.

    rewrite (ktf_prop2_lt (k := k)); try omega.
    rewrite ktf_rel2_eq_0; auto; try omega.
    apply afs_owft_afs_t with (nts_seq s n), HR1; omega.
   
    rewrite H, ktf_rel2_eq_2; auto; omega.

    (* now Hig is proof that emb_ntree R' is af on wfptree P' *)
    
    apply afs_t_impl_af_t in Hig.
    apply af_t_impl_afs_t; auto.
    revert Hig; apply ktf_ev2_af_t with (H_m_lt := H_m_lt); auto; omega.
    
  (* s (S i) = Some (in_wft_sup f), then R (S i) rlift x is secured by f x *)
    
    set (P' := @ktf_prop1 X k P i).
    assert (G0 : afs_t (P' i) (ktf_rel1_i k i lt R)).
      apply afs_t_ktf_rel1_i; auto; try omega.
      generalize (HR1 _ (le_min_r i k)); apply afs_owft_afs_t.
      generalize (HR1 _ (le_refl _)); apply afs_owft_afs_t.
    apply afs_t_afs_owft in G0.
    destruct G0 as (owfti & Howfti).
    assert (Hki : k <= S i).
      omega.
    set (R' := @ktf_rel1 _ k i m lt R).
    set (s' j := match (ktf_pos1 k i j) with
                   | inleft (inleft (inleft (left _)))   => nts_seq s j
                   | inleft (inleft (inleft (right _)))  => nts_seq s k
                   | inleft (inleft (inright _))         => owfti 
                   | inleft (inright _)                  => Some (f m)
                   | inright _                           => None
                 end).
    assert (Hs' : s' (S i) <> None /\ forall j, S i < j -> s' j = None).
      unfold s'.
      split.
      destruct (ktf_pos1 k i (S i)) as [[[[|]|]|]|]; try omega.
      discriminate.
      intros j Hj.
      destruct (ktf_pos1 k i j) as [[[[|]|]|]|]; try omega; auto.
    set (s1 := existT _ s' (exist _ (S i) Hs') : nt_stump T).    
    specialize (IHs s1); spec_all IHs; auto.

    (* s' is more facile than s *)
    
    red; right; split.
    simpl; omega.
    exists k; split.
    simpl.
    unfold s'.
    destruct (ktf_pos1 k i k) as [[[[|]|]|]|]; try omega.
    rewrite H_k_i, Hs0; apply in_wft_lt.
    intros j Hj.
    unfold s1; simpl.
    unfold s'.
    destruct (ktf_pos1 k i j) as [[[[|]|]|]|]; try omega.
    rewrite nts_prop2; auto; omega.
    
    specialize (IHs (S i) eq_refl P'); spec_all IHs; auto.
    
    apply ktf_prop1_u; auto.
    apply ktf_prop1_dec; auto.
    intros n Hn.
    unfold P'.
    do 2 (rewrite ktf_prop1_neq; try omega).
    rewrite <- (HP3 n); try omega.
    rewrite <- HP3; try omega.
    auto.
    
    specialize (IHs R'); spec_all IHs; auto.

    (* we show that s' secures P'/R' *)
    
    intros n Hn.
    simpl; unfold s', P', R'.
    destruct (ktf_pos1 k i n) as [[[[(?&?)|]|]|]|]; try omega.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_0; auto.
    apply HR1; omega.
    rewrite ktf_rel1_eq_2; auto.
    subst n; apply Howfti.
    rewrite ktf_prop1_neq; try omega.
    rewrite ktf_rel1_eq_3; auto.
    simpl; rewrite <- HP3; try omega.
    rewrite H_k_i; apply HRSi; auto.

    (* now IHs is proof that emb_tree_upto (S i) R' is af on wfptree P' *)
   
    apply afs_t_impl_af_t in IHs.
    apply af_t_impl_afs_t; auto.
    revert IHs; apply ktf_ev1_af_t with (H_m_lt := H_m_lt); auto.

(* case S i < k *)
    
    spec_all HRSi.
    omega.
    
    case_eq (nts_seq s (S i)); [ intros [ | f ] | ]; intros Hs0;
      rewrite Hs0 in HRSi; simpl in HRSi.
    Focus 3.

  (* s (S i) = None, is not possible because m belongs to P (S i) *)
           
    contradict Hm1; auto.
    
  (* s (S i) = Some in_wft_zt, then R (S i) is full on P (S i) *)
    
    set (P' := @ktf_prop4 X _ P _ H_k_i).
    assert (G0 : afs_t (P' i) (ktf_rel4_i k i lt R)).
      apply afs_t_ktf_rel4_i; auto.
      specialize (HR1 i); spec_all HR1.
      omega.
      generalize HR1; apply afs_owft_afs_t.
    apply afs_t_afs_owft in G0.
    destruct G0 as (owfti & Howfti). 
    set (R' := @ktf_rel4 _ _ _ H_k_i lt R).
    set (s' j := match (ktf_pos4 H_k_i j) with
                   | inleft (inleft (left _))   => nts_seq s j
                   | inleft (inleft (right _))  => owfti
                   | inleft (inright _)         => None 
                   | inright _                  => nts_seq s j
                 end).
    assert (Hs' : s' k <> None /\ forall j, k < j -> s' j = None).
      unfold s'.
      split.
      destruct (ktf_pos4 H_k_i k) as [[[|]|]|]; try omega.
      rewrite Hk.
      apply nts_prop1.
      intros j Hj.
      destruct (ktf_pos4 H_k_i j) as [[[|]|]|]; try omega; auto.
      rewrite nts_prop2; auto.
      rewrite <- Hk; auto.
    set (s1 := existT _ s' (exist _ _ Hs') : nt_stump T).    
    specialize (IHs s1); spec_all IHs; auto.

    (* s' is more facile than s *)
    
    red; right. 
    rewrite <- Hk.
    split; auto.
    exists (S i); split.
    rewrite Hs0.
    unfold s1; simpl.
    unfold s'.
    destruct (ktf_pos4 H_k_i (S i)) as [[[|]|]|]; try omega.
    red; constructor.
    intros j Hj.
    simpl; unfold s'.
    destruct (ktf_pos4 H_k_i j) as [[[|]|]|]; try omega; auto.
    
    specialize (IHs _ eq_refl P'); spec_all IHs; auto.
    
    apply ktf_prop4_u; auto.
    apply ktf_prop4_dec; auto.
    intros n Hn.
    unfold P'; simpl in Hn |- *.
    do 2 (rewrite ktf_prop4_neq; try omega).
    apply HP3; auto.
    
    specialize (IHs R'); spec_all IHs; auto.

    (* we show that s' secures P'/R' *)
    
    intros n Hn.
    simpl; unfold s', P', R'.
    destruct (ktf_pos4 H_k_i n) as [[[|]|]|]; try omega.
    rewrite ktf_prop4_neq; try omega.
    rewrite ktf_rel4_eq_0; auto.
    rewrite ktf_rel4_eq_1; auto.
    subst n; apply Howfti.
    rewrite ktf_rel4_eq_2; auto.
    rewrite ktf_prop4_eqS; try omega.
    red; tauto.
    rewrite ktf_prop4_neq; try omega.
    rewrite ktf_rel4_eq_3; auto.
    simpl in IHs.

    (* now IHs is proof that emb_tree_upto k R' is af on wfptree P' *)
    
    apply afs_t_impl_af_t in IHs.
    apply af_t_impl_afs_t; auto.
    revert IHs; apply ktf_ev4_af_t with (H_m_lt := H_m_lt); auto.

  (* s (S i) = Some (in_wft_sup f), then R (S i) rlift x is secured by f x *)
    
    set (P' := @ktf_prop3 X P i).
    assert (G0 : afs_t (P' i) (ktf_rel3_i k i lt R)).
      apply afs_t_ktf_rel3_i; auto.
      specialize (HR1 i); spec_all HR1.
      omega.
      generalize HR1; apply afs_owft_afs_t.
      specialize (HR1 (S i)); spec_all HR1.
      omega.
      generalize HR1; apply afs_owft_afs_t.
    apply afs_t_afs_owft in G0.
    destruct G0 as (owfti & Howfti).
    set (R' := @ktf_rel3 _ _ _ H_k_i m lt R).
    set (s' j := match (ktf_pos3 H_k_i j) with
                   | inleft (inleft (left _))   => nts_seq s j
                   | inleft (inleft (right _))  => owfti
                   | inleft (inright _)         => Some (f m) 
                   | inright _                  => nts_seq s j
                 end).
    assert (Hs' : s' k <> None /\ forall j, k < j -> s' j = None).
      unfold s'.
      split.
      destruct (ktf_pos3 H_k_i k) as [[[|]|]|]; try omega.
      rewrite Hk.
      apply nts_prop1.
      intros j Hj.
      destruct (ktf_pos3 H_k_i j) as [[[|]|]|]; try omega; auto.
      rewrite nts_prop2; auto.
      rewrite <- Hk; auto.
    set (s1 := existT _ s' (exist _ _ Hs') : nt_stump T).    
    specialize (IHs s1); spec_all IHs; auto.

    (* s' is more facile than s *)
    
    red; right. 
    rewrite <- Hk.
    split; auto.
    exists (S i); split.
    rewrite Hs0.
    unfold s1; simpl.
    unfold s'.
    destruct (ktf_pos3 H_k_i (S i)) as [[[|]|]|]; try omega.
    red; apply in_wft_lt.
    intros j Hj.
    simpl; unfold s'.
    destruct (ktf_pos3 H_k_i j) as [[[|]|]|]; try omega; auto.
    
    specialize (IHs _ eq_refl P'); spec_all IHs; auto.
    
    apply ktf_prop3_u; auto.
    apply ktf_prop3_dec; auto.
    intros n Hn.
    unfold P'; simpl in Hn |- *.
    do 2 (rewrite ktf_prop3_neq; try omega).
    apply HP3; auto.
    
    specialize (IHs R'); spec_all IHs; auto.

    (* we show that s' secures P'/R' *)
    
    intros n Hn.
    simpl; unfold s', P', R'.
    destruct (ktf_pos3 H_k_i n) as [[[|]|]|]; try omega.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_0; auto.
    rewrite ktf_rel3_eq_1; auto.
    subst n; apply Howfti.
    rewrite ktf_rel3_eq_2; auto.
    rewrite ktf_prop3_neq; try omega.
    subst n.
    apply HRSi; auto.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_3; auto.
    simpl in IHs.

    (* now IHs is proof that emb_tree_upto (S i) R' is af on wfptree P' *)
    
    apply afs_t_impl_af_t in IHs.
    apply af_t_impl_afs_t; auto.
    revert IHs; apply ktf_ev3_af_t with (H_m_lt := H_m_lt); auto.
  Qed.
  
End kruskal_ktree_rec.

Section kruskal_wfptree.
  
  Variable (X : Type) (k : nat) (P : nat -> X -> Prop) (R : nat -> X -> X -> Prop).
  Hypothesis HP1 : forall n x, { P n x } + { ~ P n x }. 
  Hypothesis HP2 : forall n, k <= n -> P k = P n.
  Hypothesis HR  : forall n, n <= k -> afs_t (P n) (R n).
  
  Theorem kruskal_wfptree : afs_t (wfptree P) (emb_tree_upto k R).
  Proof.

    set (P' n (t : ktree_fix X) := 
      match t with
        | in_ktf_0 x => P n x 
        | _          => False
      end).

    assert (forall n x, { P' n x } + { ~ P' n x }) as Pdec'.
      intros n []; unfold P'; try tauto; auto.

    set (R' n (t1 t2 : ktree_fix X) :=
      match t1, t2 with
        | in_ktf_0 u, in_ktf_0 v => R n u v
        | _         , _          => False
      end). 

    assert (HR' : forall n, n <= k -> afs_t (P' n) (R' n)).
      intros n Hn; generalize (HR Hn).
      apply afs_t_relmap with (f := fun x y => y = in_ktf_0 x); auto.
      unfold P'; intros [ | x | | ] H; try (destruct H; fail). 
      exists x; auto.
      unfold P'; intros x1 x2 [ ] [ ] H1 H2 H3 H4 H7 H8; try tauto.
      injection H7; injection H8; intros ? ?; subst; auto.
                                                
    set (s1 n := match le_lt_dec n k with
                   | left H  => Some (proj1_sig (HR' _ H))
                   | right _ => None
                 end).

    assert (Hs1 :  s1 k <> None /\ forall i, k < i -> s1 i = None).
      unfold s1; split.
      destruct (le_lt_dec k k); try omega; discriminate.
      intros i Hi.
      destruct (le_lt_dec i k); try omega; auto.
      
    set (s := existT _ s1 (exist _ k Hs1) : nt_stump _ ).
    
    generalize (kruskal_ktree_rec s eq_refl P'); intros K; spec_all K; auto.
    
    intros n; unfold nts_char; simpl; intros Hn.
    unfold P'.
    rewrite <- (HP2 Hn); auto.
    
    specialize (K R'); spec_all K.

    unfold s; simpl.
    intros n Hn.
    unfold s1.
    destruct (le_lt_dec n k) as [ H | ]; try omega; simpl.
    apply (proj2_sig (HR' _ H)).
    
    revert K.
    simpl nts_char.
    clear HR' s1 Hs1 s.
    
    apply afs_t_relmap with (f := fun x y => x = tree_map (@in_ktf_0 _) y).

    apply wfptree_dec; auto.

    intros y Hy; exists (tree_map (@in_ktf_0 _) y); split; auto.
    revert Hy.
    induction 1 as [ a ll Hll IH ] using wfptree_ind.
    rewrite tree_map_fix.
    revert Hll; do 2 rewrite wfptree_fix.
    intros (G1 & G2); split.
    rewrite map_length.
    red; auto.
    intros t Ht.
    apply in_map_iff in Ht.
    destruct Ht as (u & ? & Ht); subst t; auto.
    
    intros x1 x2 y1 y2 _ _ _ _ G1 G2 H.
    revert H y1 y2 G1 G2.    
    induction 1 as [ t1 t2 x ll Hll H IH 
                   | x y ll mm Hll Ha Hb IH
                   | x y ll mm Hll Ha Hb IH 
                   ] using emb_tree_upto_ind.
                   
    intros y1 [ b lb ] ? H1; subst t1.
    rewrite tree_map_fix in H1.
    injection H1; clear H1; intros ? ?; subst.
    rewrite in_map_iff in Hll.
    destruct Hll as (u & ? & Hll); subst t2.
    apply in_emb_tree_upto_0 with u; auto.
    
    intros [ c kk ] [ d pp ] G4 G5.
    rewrite tree_map_fix in G4.
    rewrite tree_map_fix in G5.
    injection G4; injection G5.
    clear G4 G5; intros ? ? ? ?; subst.
    rewrite map_length in Hll.
    rewrite map_length in Ha.
    red in Ha.
    apply in_emb_tree_upto_1; auto.
    rewrite Forall2_map_both in IH.
    revert IH.
    apply Forall2_impl.
    intros ? ? H; apply H; auto.
    
    intros [ c kk ] [ d pp ] G4 G5.
    rewrite tree_map_fix in G4.
    rewrite tree_map_fix in G5.
    injection G4; injection G5;
    clear G4 G5; intros ? ? ? ?; subst.
    rewrite map_length in Hll.
    red in Ha.
    apply in_emb_tree_upto_2; auto.
    apply subword_map_both in IH.
    revert IH.
    apply subword_inc.
    intros ? ? H; apply H; auto.
  Qed.

End kruskal_wfptree.


