(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Permutation.

Require Omega.

Require Import list_perm.

Set Implicit Arguments.

(* list product *)

Section list_prod.

  Variables (A B C : Type) (f : A -> B -> C).

  (* list_prod [a1;...;an] [b1;...;bm] = [f a1 b1; f a1 b2;...;f a1 bm;f a2 b1;...;f a2 bm;...;f an b1;f an bm] *)

  Definition list_prod ll mm := flat_map (fun x => map (f x) mm) ll.

  Fact list_prod_spec ll mm : forall c, In c (list_prod ll mm) <-> exists a b, In a ll /\ In b mm /\ c = f a b.
  Proof.
    intros c; split; unfold list_prod;
    rewrite in_flat_map.
    intros (a & H1 & H2).
    rewrite in_map_iff in H2.
    destruct H2 as (b & H2 & H3).
    exists a, b; auto.
    intros (a & b & H1 & H2 & H3).
    exists a; split; auto.
    rewrite in_map_iff.
    exists b; auto.
  Qed.

  Fact list_prod_length ll mm : length (list_prod ll mm) = length ll * length mm.
  Proof.
    induction ll; simpl; auto; rewrite app_length, map_length; f_equal; auto.
  Qed.   
  
  Fact list_prod_nil ll mm : list_prod ll mm = nil -> ll = nil \/ mm = nil.
  Proof.
    intros H.
    apply f_equal with (f := @length _) in H.
    rewrite list_prod_length in H; simpl in H.
    destruct ll; auto.
    destruct mm; auto.
    simpl in H; discriminate H.
  Qed.    

  Fact list_prod_nil_right ll : list_prod ll nil = nil.
  Proof.
    unfold list_prod.
    induction ll; simpl; auto.
  Qed.

  Fact list_prod_sg_right ll x : list_prod ll (x::nil) = map (fun a => f a x) ll.
  Proof.
    induction ll; simpl; auto.
  Qed.

  Fact list_prod_app ll mm nn : list_prod ll (mm++nn) ~p list_prod ll mm ++ list_prod ll nn.
  Proof.
    induction ll as [ | x ll IHll ]; simpl.
    apply Permutation_refl.
    rewrite map_app;
    repeat rewrite <- app_assoc.
    apply Permutation_app.
    apply Permutation_refl.
    apply Permutation_trans with (2 := Permutation_app_comm _ _).
    rewrite <- app_assoc.
    apply Permutation_app.
    apply Permutation_refl.
    apply Permutation_trans with (1 := IHll).
    apply Permutation_app_comm.
  Qed.

  Fact list_prod_cons ll a mm : list_prod ll (a::mm) ~p list_prod ll (a::nil) ++ list_prod ll mm.
  Proof.
    apply (list_prod_app _ (a::nil) mm).
  Qed.
  
  Let list_prod_perm_left ll mm nn : ll ~p mm -> list_prod ll nn ~p list_prod mm nn.
  Proof.
    intros H; revert nn; induction H; intros nn; simpl; auto.
    apply Permutation_app; auto.
    do 2 rewrite <- app_ass.
    apply Permutation_app; auto.
    apply Permutation_app_comm.
    eapply Permutation_trans; auto.
  Qed.
  
  Let list_prod_perm_right nn ll mm : ll ~p mm -> list_prod nn ll ~p list_prod nn mm.
  Proof.
    intros H; revert nn; induction H; intros nn; simpl; auto.
    do 2 (apply Permutation_trans with (1 := list_prod_cons _ _ _);
          rewrite list_prod_sg_right;
          apply Permutation_sym).
    apply Permutation_app; auto.
    do 2 (apply Permutation_trans with (1 := list_prod_cons _ _ _);
          rewrite list_prod_sg_right;
          apply Permutation_sym).
    generalize (list_prod_cons nn x l) (list_prod_cons nn y l); intros H1 H2.
    apply Permutation_app with (1 := Permutation_refl (map (fun a : A => f a y) nn)) in H1.
    apply Permutation_app with (1 := Permutation_refl (map (fun a : A => f a x) nn)) in H2.
    apply Permutation_trans with (1 := H1),
          Permutation_sym,
          Permutation_trans with (1 := H2),
          Permutation_sym.
    do 2 rewrite list_prod_sg_right.
    do 2 rewrite <- app_ass.
    apply Permutation_app; auto.
    apply Permutation_app_comm.
    eapply Permutation_trans; auto.
  Qed.

  (* list_prod is congruent under permutations *)

  Fact list_prod_perm l1 l2 m1 m2 : l1 ~p l2 -> m1 ~p m2 -> list_prod l1 m1 ~p list_prod l2 m2.
  Proof.
    intros H1 H2.
    apply Permutation_trans with (1 := list_prod_perm_left _ H1).
    apply list_prod_perm_right; auto.
  Qed.

End list_prod.
