(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Permutation.
Require Import Relations.

Require Import rel_utils.
Require Import list_utils.
Require Import continuity.

Set Implicit Arguments.

Section perm_R.

  Variable (X : Type) (R : X -> X -> Prop).

  Definition permR ll mm := exists ii, Permutation ll ii /\ Forall2 R ii mm.

  Infix "~r" := permR (at level 70).

  Fact permR_refl ll : Forall (fun x => R x x) ll -> permR ll ll.
  Proof.
    induction 1 as [ | x ll H IH ].
    exists nil; auto.
    exists (x::ll); split; auto.
    rewrite <- Forall_Forall2.
    constructor; auto.
  Qed.

  Fact permR_sym : symmetric _ R -> symmetric _ permR.
  Proof.
    intros H ll mm (zz & H1 & H2).
    apply Permutation_sym in H1.
    apply Forall2_sym in H2; auto.
    apply Forall2_perm_comm with (1 := H2) (2 := H1).
  Qed.

  Fact permR_trans : transitive _ R -> transitive _ permR.
  Proof.
    intros HR x y z (ii & H4 & H5) (jj & H6 & H7).
    destruct Forall2_perm_comm with (1 := H5) (2 := H6) as (qq & H8 & H9).
    exists qq; split.
    apply Permutation_trans with (1 := H4); auto.
    apply Forall2_trans with (3 := H7); auto.
  Qed.

  Fact permR_In s ll kk : In s ll -> kk ~r ll -> exists s', In s' kk /\ R s' s.
  Proof.
    intros H1 (ii & H2 & H3).
    apply in_split in H1.
    destruct H1 as (l & r & H1); subst ll.
    apply Forall2_app_inv_r in H3.
    destruct H3 as (l' & [ | s' r' ] & H3 & H4 & H5); subst;
    inversion_clear H4.
    exists s'; split; auto.
    apply Permutation_sym in H2. 
    apply Permutation_in with (1 := H2).
    apply in_app_iff; right; left; auto.
  Qed.

  Fact permR_In_inv l a r ll  : l++a::r ~r ll -> exists l' a' r', ll = l'++a'::r' /\ R a a' /\ l++r ~r l'++r'.
  Proof.
    intros (ii & H1 & H2).
    destruct Permutation_In_inv with (x := a) (1 := H1) as (l1 & r1 & ?); subst.
    apply in_or_app; right; left; auto.
    apply Forall2_app_inv_l in H2.
    destruct H2 as (l' & r2 & H2 & H3 & ?); subst.
    inversion_clear H3.
    exists l', y, l'0; repeat split; auto.
    exists (l1++r1); split.
    apply Permutation_app_inv with (1 := H1).
    apply Forall2_app; auto.
  Qed.

  Fact permR_add_middle l1 m1 r1 l2 m2 r2 : l1++r1 ~r l2++r2 -> m1 ~r m2 -> l1++m1++r1 ~r l2++m2++r2.
  Proof.
    intros (lr & H1 & H2) (ii & H3 & H4).
    apply Forall2_app_inv_r in H2.
    destruct H2 as (l & r & H2 & H5 & H6).
    subst.
    exists (l++ii++r); split.
    apply Permutation_app_intro with (k := m1) in H1.
    apply Permutation_trans with (1 := H1).
    repeat apply Permutation_app; auto.
    repeat apply Forall2_app; auto.
  Qed.

End perm_R.

Fact permR_monotonic X : op_monotonic (@permR X).
Proof.
  intros R S HRS x y (z & H1 & H2).
  exists z; split; auto.
  clear H1; induction H2; auto.
Qed.

Fact permR_w_continuous X : op_w_continuous (@permR X).
Proof.
  intros K HK x y (z & H1 & H2).
  apply list_Forall2_sup in H2; auto.
  destruct H2 as (n & H2).
  exists n, z; split; auto.
Qed.
