(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.
Require Import tree.

Set Implicit Arguments.

Section emb_tree_product.

  Variables (X : Type) (R : X -> X -> Prop). 
  
  Reserved Notation "x '<et' y" (at level 70, no associativity).
  
  (** This is the "neatly embeds" definition of page 237 of W. Veldman 
   ** but the relation R does not depend on the arity of nodes
   **
   ** U=<x|u1,...,un> neatly embeds into V=<y|v1,...,vp> iff
   **    a/ U neatly embeds into vi for some i in {1,...,p}
   ** or b/ R x y and n = p and ui neatly embeds into vi for all i in {1,...,n=p}
   **)

  Unset Elimination Schemes.
  
  Inductive emb_tree_product : tree X -> tree X -> Prop :=
    | in_emb_tree_product_0 : forall s t x ll, In t ll -> s <et t -> s <et in_tree x ll
    | in_emb_tree_product_1 : forall x y ll mm, R x y -> Forall2 emb_tree_product ll mm -> in_tree x ll <et in_tree y mm
  where "x <et y" := (emb_tree_product x y).
  
  Set Elimination Schemes.
    
  Section emb_tree_product_ind.
  
    Variable S : tree X -> tree X -> Prop.
    
    Infix "<<" := S (at level 70).
    
    Hypothesis S_sub0 : forall s t x ll, In t ll -> s <et t -> s << t -> s << in_tree x ll.
    Hypothesis S_sub1 : forall x y ll mm, R x y 
                                       -> Forall2 emb_tree_product ll mm 
                                       -> Forall2 S                ll mm 
                                       -> in_tree x ll << in_tree y mm.
                                       
    Fixpoint emb_tree_product_ind t1 t2 (H : t1 <et t2) : t1 << t2.
    Proof.
      destruct H as [ s t x ll H1 H2 | x y ll mm H1 H2 ].
      
      apply S_sub0 with t; auto.
      apply S_sub1; auto.

      revert H2; clear x y H1.
      induction 1; constructor; auto.
    Qed.
      
  End emb_tree_product_ind.
  
  Fact emb_sub_tree_product r s t : r <et s -> s <st t -> r <et t.
  Proof.
    intros Hrs.
    induction 1 as [ | ? ? ? ? H ]; auto.
    constructor 1 with (1 := H); auto.
  Qed.
  
  Fact sub_emb_tree_product r s t : r <st s -> s <et t -> r <et t.
  Proof.
    intros H.
    induction 1 as [ s t x ll H1 H2 IH2 
                   | x y ll mm H1 H2 IH2 ].
    constructor 1 with (1 := H1); auto.
    apply sub_tree_inv in H.
    destruct H as [ H | H ].
    subst; constructor 2; auto.
    
    apply ssub_tree_inv_right in H.
    destruct H as (s & H3 & z & kk & H4 & H5).
    injection H5; clear H5; intros ? ?; subst z kk.
    apply in_split in H4; destruct H4 as (u & v & H4); subst ll.
    apply Forall2_app_inv_l in IH2.
    destruct IH2 as (c & d & _ & IH2 & ?); subst.
    apply Forall2_cons_inv_l in IH2.
    destruct IH2 as (t & e & IH2 & ? & _); subst.
    constructor 1 with t; auto.
    apply in_or_app; right; simpl; auto.
  Qed.
  
  Fact sub_emb_tree_product_leaf_inv x y t : R x y -> in_tree y nil <st t -> in_tree x nil <et t.
  Proof.
    revert x.
    induction t as [ x2 ll IH ]; intros x1 Hx1.
    intros H.
    apply sub_in_tree_inv in H.
    destruct H as [ H | (? & H1 & H2) ].
    
    injection H; clear H; intros; subst.
    constructor 2; auto.
    
    constructor 1 with x; auto.
  Qed.

End emb_tree_product.

Section emb_tree_higman.

  Variables (X : Type) (R : nat -> X -> X -> Prop). 
    
  Reserved Notation "x '<et' y" (at level 70, no associativity).

  (** Same as emb_tree_product but there is one relation for each arity 
      We use that one in the inductive proof of the Higman theorem
   **)

  Unset Elimination Schemes.
  
  Inductive emb_tree_higman : tree X -> tree X -> Prop :=
    | in_embnt_0 : forall s t x ll, In t ll -> s <et t -> s <et in_tree x ll
    | in_embnt_1 : forall x y ll mm, R (length ll) x y -> Forall2 emb_tree_higman ll mm -> in_tree x ll <et in_tree y mm
  where "x <et y" := (emb_tree_higman x y).
  
  Set Elimination Schemes.
  
  Section emb_tree_higman_ind.
  
    Variable S : tree X -> tree X -> Prop.
    
    Infix "<<" := S (at level 70).
    
    Hypothesis S_sub0 : forall s t x ll, In t ll -> s <et t -> s << t -> s << in_tree x ll.
    Hypothesis S_sub1 : forall x y ll mm, R (length ll) x y 
                                       -> Forall2 emb_tree_higman ll mm 
                                       -> Forall2 S               ll mm 
                                       -> in_tree x ll << in_tree y mm.
                                       
    Fixpoint emb_tree_higman_ind t1 t2 (H : t1 <et t2) : t1 << t2.
    Proof.
      destruct H as [ s t x ll H1 H2 | x y ll mm H1 H2 ].
      
      apply S_sub0 with t; auto.
      apply S_sub1; auto.
      revert H2; clear x y H1.
      induction 1; constructor; auto.
    Qed.
      
  End emb_tree_higman_ind.
  
  Fact emb_sub_tree_higman r s t : r <et s -> s <st t -> r <et t.
  Proof.
    intros Hrs.
    induction 1 as [ | ? ? ? ? H ]; auto.
    constructor 1 with (1 := H); auto.
  Qed.
    
  Fact sub_emb_tree_higman r s t : r <st s -> s <et t -> r <et t.
  Proof.
    intros H.
    induction 1 as [ s t x ll H1 H2 IH2 
                   | x y ll mm H1 H2 IH2 ].
    constructor 1 with (1 := H1); auto.
    apply sub_tree_inv in H.
    destruct H as [ H | H ].
    subst; constructor 2; auto.
    
    apply ssub_tree_inv_right in H.
    destruct H as (s & H3 & z & kk & H4 & H5).
    injection H5; clear H5; intros ? ?; subst z kk.
    apply in_split in H4; destruct H4 as (u & v & H4); subst ll.
    apply Forall2_app_inv_l in IH2.
    destruct IH2 as (c & d & _ & IH2 & ?); subst.
    apply Forall2_cons_inv_l in IH2.
    destruct IH2 as (t & e & IH2 & ? & _); subst.
    constructor 1 with t; auto.
    apply in_or_app; right; simpl; auto.
  Qed.
  
  Fact sub_emb_tree_higman_leaf_inv x y t : R 0 x y -> in_tree y nil <st t -> in_tree x nil <et t.
  Proof.
    revert x.
    induction t as [ x2 ll IH ]; intros x1 Hx1.
    intros H.
    apply sub_in_tree_inv in H.
    destruct H as [ H | (? & H1 & H2) ].
    
    injection H; clear H; intros; subst.
    constructor 2; auto.

    constructor 1 with x; auto.
  Qed.

End emb_tree_higman.

Fact emb_tree_product_higman_eq X R : emb_tree_higman (fun _ => R) ~eq2 @emb_tree_product X R.
Proof.
  split; intros t1 t2.
  induction 1 as [ ? t | ].
  constructor 1 with t; auto.
  constructor 2; auto.
  induction 1 as [ ? t | ].
  constructor 1 with t; auto.
  constructor 2; auto.
Qed.

Section emb_tree_higman_inc.

  Variable (X : Type) (R S : nat -> X -> X -> Prop).
  
  Hypothesis HRS : forall n, R n inc2 S n.
  
  Fact emb_tree_higman_inc : emb_tree_higman R inc2 emb_tree_higman S.
  Proof.
    induction 1 as [ s t x ll H1 H2 IH2 
                   | x y ll mm H1 H2 IH2 ].
    constructor 1 with t; auto.
    constructor 2; auto.
  Qed.

End emb_tree_higman_inc.

Section emb_wfptree_higman_ind.

  Variables (X : Type) (P : nat -> X -> Prop) (R : nat -> X -> X -> Prop).
  
  Infix "<et" := (emb_tree_higman R) (at level 70, no associativity).
  
  Variable S : tree X -> tree X -> Prop.
  
  Infix "<<" := S (at level 70).
  
  Hypothesis S_sub0 : forall t t' a ll, wfptree P t 
                                     -> wfptree P (in_tree a ll)
                                     -> In t' ll 
                                     -> t <et t' 
                                     -> t << t' 
                                     -> t << in_tree a ll.
                                     
  Hypothesis S_sub1 : forall a b ll mm, wfptree P (in_tree a ll)
                                     -> wfptree P (in_tree b mm) 
                                     -> R (length ll) a b 
                                     -> Forall2 (emb_tree_higman R) ll mm 
                                     -> Forall2 S ll mm 
                                     -> in_tree a ll << in_tree b mm.

  Theorem emb_wfptree_higman_ind t1 t2 : wfptree P t1 -> wfptree P t2 -> t1 <et t2 -> t1 << t2.
  Proof.
    intros H1 H2 H; revert H H1 H2.
    induction 1 as [ s t x ll Hll H IH 
                   | x y ll mm Hll Ha IH
                   ] using emb_tree_higman_ind; intros H1 H2.
                   
    generalize H2; rewrite wfptree_fix; intros (? & ?). 
    apply S_sub0 with t; auto.
    
    apply S_sub1; auto.
    apply wfptree_fix, proj2, Forall_forall in H1.
    apply wfptree_fix, proj2, Forall_forall in H2.
    apply Forall2_Forall_left with (1 := IH) in H1.
    apply Forall2_Forall_right with (1 := H1); auto.
  Qed.

End emb_wfptree_higman_ind.
