(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import perm_eq.
Require Import continuity.

Require Import tree.

Set Implicit Arguments.

Section tree_eq.

  (* We define tree_eq : tree -> tree -> Prop by

     tree_eq (in_tree f ll) (in_tree g mm) 
 
             iff 

     f = g and ll is a permutation of mm upto tree_eq 

     Hence we have a nested inductive predicate. 
     We define it using a omega-continuous LFP which 
     gives us easy proofs of refl, sym and trans.

  *)

  Variable X : Type.

  Let fop (S : relation (tree X)) := fun t1 t2 =>
    match t1        , t2 with
      | in_tree f ll, in_tree g mm => f = g /\ permR S ll mm
    end.

  Hint Resolve permR_monotonic permR_w_continuous.

  Let fop_mono : op_monotonic fop.
  Proof.
    intros R S HRS [] [] [ H1 H2 ]; split; auto.
    revert H2; apply permR_monotonic; auto.
  Qed.

  Let fop_w_continuous : op_w_continuous fop.
  Proof.
    intros K HK [] [] [ ? H ]; subst.
    rewrite op_mono_w_cont_eq in H; auto.
    destruct H as (n & ?); exists n; red; auto.
  Qed.

  Definition tree_eq := lfp fop.

  Infix "~t" := tree_eq (at level 70).
  Infix "~p" := (permR tree_eq) (at level 70).

  Fact tree_eq_fixpoint_eq f g ll mm : in_tree f ll ~t in_tree g mm <-> f = g /\ ll ~p mm.
  Proof.
    unfold tree_eq.
    rewrite lfp_fixpoint_eq; auto.
    unfold fop at 1; split; auto.
  Qed.

  Fact in_tree_eq : forall f ll mm, ll ~p mm -> in_tree f ll ~t in_tree f mm.
  Proof.  
    intros; rewrite tree_eq_fixpoint_eq; auto; split; auto.
  Qed.

  Fact tree_eq_inv f g ll mm : in_tree f ll ~t in_tree g mm -> f = g /\ ll ~p mm.
  Proof.
    apply tree_eq_fixpoint_eq.
  Qed.

  Section tree_eq_ind.
    
    Variable R : tree X -> tree X -> Prop.
    
    Hypothesis HR : forall f ll mm, ll ~p mm -> permR R ll mm -> R (in_tree f ll) (in_tree f mm).
    
    Theorem tree_eq_ind : tree_eq inc2 R.
    Proof.
      apply lfp_ind; auto.
      intros [ f ll ] [ g mm ] (? & H); subst.
      apply HR; revert H; apply permR_monotonic; intros; tauto.
    Qed.

  End tree_eq_ind.     

  Section tree_eq_least.

    Variable R : tree X -> tree X -> Prop.
    
    Hypothesis HR : forall f ll mm, permR R ll mm -> R (in_tree f ll) (in_tree f mm).
    
    Theorem tree_eq_induction t1 t2 : t1 ~t t2 -> R t1 t2.
    Proof.
      apply tree_eq_ind.
      intros; apply HR; auto.
    Qed.

  End tree_eq_least.

  Theorem tree_eq_refl : reflexive _ tree_eq.
  Proof.
    unfold tree_eq.
    intros t; induction t as [ f ll IH ].
    rewrite lfp_fixpoint_eq; auto; split; auto.
    apply permR_refl; auto.
    rewrite Forall_forall; auto.
  Qed.

  Theorem tree_eq_sym : symmetric _ tree_eq.
  Proof.
    unfold tree_eq.
    apply lfp_induction; auto.
    intros ? ? [].
    intros S HS [f ll] [g mm] [H1 H2]; subst; split; auto.
    revert H2; apply permR_sym; auto.
    intros ? ? H ? ? (n & ?); exists n; apply H; auto.
  Qed.

  Theorem tree_eq_trans : transitive _ tree_eq.
  Proof.
    unfold tree_eq.
    apply lfp_induction; auto.
    intros ? ? ? [].
    intros S HS [f ll] [g mm] [h kk] (H1 & H2) (H3 & H4); subst; split; auto.
    revert H2 H4; apply permR_trans; auto.
    intros K HK H ? ? ? (n1 & H1) (n2 & H2); exists (n1+n2).
    apply (H (n1+n2) x y z).
    revert H1; apply seq_inc_all with (1 := HK), le_plus_l.
    revert H2; apply seq_inc_all with (1 := HK), le_plus_r.
  Qed.

  Hint Resolve tree_eq_refl tree_eq_sym tree_eq_trans.

  Fact permR_tree_eq_refl : reflexive _ (permR tree_eq).
  Proof. 
    intro; apply permR_refl.
    rewrite Forall_forall; intros; auto.
  Qed.
 
  Fact permR_tree_eq_sym : symmetric _ (permR tree_eq).
  Proof. apply permR_sym; auto. Qed.

  Fact permR_tree_eq_trans : transitive _ (permR tree_eq).
  Proof. apply permR_trans; auto. Qed.

End tree_eq.

