(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.
Require Import Bool.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import sublist.
Require Import subword.
Require Import intercalate.

Require Import tree.
Require Import tree_embeddings.

Require Import af_t.
Require Import afs_t.

Require Import higman_lemma.
Require Import higman_thm.

Set Implicit Arguments.

Section qtrees.

  Variable (X : Type) (R : X -> X -> Prop) (HR : forall x y, R x y).

  Variable (n : nat) (x0 : X) (ll : list (tree X)) (Hll : length ll = S n).  

  Infix "<eh" := (emb_tree_homeo R) (at level 70).

  Let Heh_dec : forall t1 t2, { t1 <eh t2 } + { ~ t1 <eh t2 }.
  Proof.
    apply emb_tree_homeo_dec; left; auto.
  Defined.  

  Let Y := (X * { k | k < n })%type.
  Let T := (X * forall t, In_t t ll -> { ll | Forall (fun x => ~ t <eh x) ll })%type.
  
  Definition vaz_ar : Y + T -> nat.
  Proof.
    intros [ (_,(k & _)) | _ ].
    exact k.
    exact n.
  Defined.
 
  Definition vaz_ev (t : tree (Y+T)) : tree X.
  Proof.
    induction t as [ [ (x,(k,Hk)) | (x,f) ] lx IH ] using tree_recursion.
    apply (in_tree x IH).
    apply (in_tree x).
    apply (intercalate IH (list_In_t_map ll (fun x Hx => proj1_sig (f x Hx)))).
  Defined.
  
  Let vaz_ev_left x k Hk lx : vaz_ev (in_tree (inl (x,exist _ k Hk)) lx) = in_tree x (map vaz_ev lx).
  Proof.
    apply tree_recursion_fix.
  Qed.
  
  Let vaz_ev_right x f lx : vaz_ev (in_tree (inr (x,f)) lx) 
                              = in_tree x (intercalate (map vaz_ev lx) (list_In_t_map ll (fun x Hx => proj1_sig (f x Hx)))).
  Proof.
    apply tree_recursion_fix.
  Qed.

  Fact vaz_ev_wf t : stree vaz_ar t -> ~ in_tree x0 ll <eh vaz_ev t.
  Proof.
    induction t as [ [ (x,(k,Hk)) | (x,f) ] lx IH ];
    intros Hlx; rewrite stree_fix in Hlx;
    simpl in Hlx; destruct Hlx as (H1 & H2); intros H.
    
    rewrite vaz_ev_left in H.
    apply emb_tree_homeo_inv in H.
    destruct H as [ (t & H3 & H4) | (H3 & H4) ].

    rewrite in_map_iff in H3.
    destruct H3 as (u & ? & H3); subst t.
    revert H4; apply IH; auto.
    
    apply subword_length in H4.
    rewrite map_length in H4.
    omega.
    
    rewrite vaz_ev_right in H.
    apply emb_tree_homeo_inv in H.
    destruct H as [ (t & H3 & H4) | (H3 & H4) ].
    
    rewrite intercalate_In_spec in H3.
    destruct H3 as [ H3 | (lw & H3 & H5) ].
    rewrite in_map_iff in H3.
    destruct H3 as (u & ? & H3); subst t.
    revert H4; apply IH; auto.
    revert H5; apply In_t_P; intros H5.
    apply In_t_list_In_t_map in H5.
    destruct H5 as (u & Hu & H5); subst lw.
    generalize (proj2_sig (f u Hu)); simpl; intros H5.
    rewrite Forall_forall in H5.
    apply H5 in H3.
    contradict H3; revert H4.
    apply sub_emb_tree_homeo.
    constructor 2 with u.
    apply In_t_In; auto.
    constructor.
    
    assert (Hf : forall x Hx y, In y (proj1_sig (f x Hx)) -> ~ x <eh y).
      intros u Hu; rewrite <- Forall_forall; apply (proj2_sig (f u Hu)).
    revert H4; apply subword_not_intercalate.
    2: rewrite map_length; omega.
    rewrite <- (map_id ll) at 1.
    rewrite <- (list_In_t_map_t (fun x => x) ll).
    apply Forall2_list_In_t_map.
    intros ? ? ?; rewrite map_t_prop; apply Hf.
  Qed.

  Fact vaz_ev_surj (y : tree X) : ~ in_tree x0 ll <eh y -> { x | vaz_ev x = y /\ stree vaz_ar x }.
  Proof.
    induction y as [ x mm IH ]; intros Hy.
    
    rewrite emb_tree_homeo_inv in Hy.
    assert (forall t, In t mm -> ~ in_tree x0 ll <eh t) as H1.
      intros t Ht; contradict Hy.
      left; exists t; auto.
    assert (~ subword (emb_tree_homeo R) ll mm) as H2.
      contradict Hy; right; auto.
    clear Hy.
    rewrite <- In_In_t in H1.
    assert (IH' : forall x, In_t x mm -> { t | vaz_ev t = x /\ stree vaz_ar t }).
      intros; apply IH; auto.
    clear IH; rename IH' into IH.
    set (f := fun x Hx => proj1_sig (IH x Hx)).
    assert (Hf1 : forall x Hx, vaz_ev (f x Hx) = x). 
      intros u Hu; apply (proj2_sig (IH u Hu)).
    assert (Hf2 : forall x Hx, stree vaz_ar (f x Hx)). 
      intros u Hu; apply (proj2_sig (IH u Hu)).
    generalize f Hf1 Hf2; clear IH f Hf1 Hf2; intros f Hf1 Hf2.
    
    destruct (le_lt_dec n (length mm)) as [ H3 | H3 ].
    
    (** case where n <= length mm 
     ** then ~ ll <sw mm implies that mm can be written
     ** mm = intercalate w lw with length w = n and 
     ** length lw = S n and every element of ll is not
     ** embedded into each element of the corresponding list in lw
     *) 
    
    apply not_subword_intercalate with (n := n) in H2; auto.
    destruct H2 as (w & lw & H2 & H4 & H5 & H6).
    assert (Hw : forall x, In_t x w -> In_t x mm).
      rewrite H2.
      intros; apply intercalate_In_t; left; auto.
    set (kk := list_In_t_map _ (fun x Hx => f _ (Hw x Hx))).
    apply Forall2_to_list_In_t_map in H6.
    destruct H6 as (g & Hg1 & Hg2).
    set (t := fun x Hx => exist _ (g x Hx) (Hg2 x Hx)).
    exists (in_tree (inr (x,t)) kk); split.
    rewrite vaz_ev_right; f_equal.
    rewrite H2; f_equal.
    unfold kk.
    rewrite map_list_In_t_map.
    rewrite <- map_id, <- list_In_t_map_t.
    apply list_In_t_map_ext; intros; rewrite Hf1, map_t_prop; auto.
    rewrite Hg1; auto.
    apply stree_fix; simpl; split.
    unfold kk; rewrite list_In_t_map_length; auto.
    intro u; apply In_t_P; unfold kk; intros Hu.
    apply In_t_list_In_t_map in Hu.
    destruct Hu as (v & Hu & ?); subst u; auto.
    
    (* case where length mm < n *)
    
    exists (in_tree (inl (x,exist _ (length mm) H3)) (list_In_t_map mm f)).
    rewrite vaz_ev_left.
    split.    
    rewrite map_list_In_t_map.
    rewrite <- (map_id mm) at 2.
    rewrite <- list_In_t_map_t.
    match goal with |- in_tree _ ?a = in_tree _ ?b => assert (a = b) as H end.
    apply list_In_t_map_ext; intros; rewrite Hf1, map_t_prop; auto.
    rewrite H; auto.
    apply stree_fix; simpl; split.
    rewrite list_In_t_map_length; auto.
    intros t; apply In_t_P; intros Ht.
    apply In_t_list_In_t_map in Ht.
    destruct Ht as (a & Ht & ?); subst t; auto.
  Defined.

  Definition vaz_rel (a b : Y + T) :=
    match a, b with
      | inl _, inl _         => True
      | inr (_,f), inr (_,g) => forall t Ht, subword (emb_tree_homeo R) (proj1_sig (f t Ht)) (proj1_sig (g t Ht))
      | _    , _             => False
    end.

  Notation R' := vaz_rel.
    
  Fact vaz_ev_emb t1 t2 : emb_tree_homeo R' t1 t2 -> stree vaz_ar t1 -> stree vaz_ar t2 -> vaz_ev t1 <eh vaz_ev t2.
  Proof.
    induction 1 as [ s t x mm H1 H2 IH2 
                   | x y kk mm H1 H2 IH2 ] using emb_tree_homeo_ind;
      repeat rewrite stree_fix.
      
    intros Hs (Hx & Hmm).
    spec_all IH2; auto.
    destruct x as [ (x,(k,Hk)) | (x,f) ].
    
    rewrite vaz_ev_left.
    apply in_emb_tree_homeo_0 with (vaz_ev t); auto.
    rewrite in_map_iff; exists t; auto.
    
    rewrite vaz_ev_right.
    apply in_emb_tree_homeo_0 with (vaz_ev t); auto.
    rewrite intercalate_In_spec.
    left; rewrite in_map_iff; exists t; auto.
    
    intros (Hx & Hkk) (Hy & Hmm).
    assert (subword (emb_tree_homeo R) (map vaz_ev kk) (map vaz_ev mm)) as IH3.
      revert IH2.
      do 2 rewrite subword_sublist_eq.
      intros (hh & H3 & H4).
      exists (map vaz_ev hh); split.
      revert H3.
      rewrite Forall2_map_both.
      rewrite <- Forall_forall in Hkk.
      assert (Forall (stree vaz_ar) hh) as Hhh.
        rewrite Forall_forall.
        intros u Hu; apply Hmm; revert Hu.
        apply sl_In; auto.
      intros H5.
      assert (length kk = length hh) as H7.
        apply Forall2_length in H5; auto.
      assert (Forall2 (fun x y => stree vaz_ar x /\ stree vaz_ar y) kk hh) as H6.
        apply Forall2_conj.
        rewrite Forall2_right_Forall, Forall2_left_Forall; auto.      
      revert H5; apply Forall2_impl_Forall2_incl.
      revert H6; apply Forall2_impl_Forall2_incl.
      apply Forall2_True; auto.
      intros; tauto.
      revert H4; apply sublist_map.
    clear IH2.
    destruct x as [ (x,(k,Hk)) | (x,f) ];
    destruct y as [ (y,(p,Hp)) | (y,g )];
    repeat rewrite vaz_ev_left; repeat rewrite vaz_ev_right.
    
    apply in_emb_tree_homeo_1; auto.
    
    apply in_emb_tree_homeo_1; auto.
    apply subword_sublist_trans with (1 := IH3).
    apply intercalate_sublist.
    
    simpl in Hy, Hx.
    apply subword_length in IH3.
    do 2 rewrite map_length in IH3.
    omega.
    
    apply in_emb_tree_homeo_1; auto.
    simpl in Hy, Hx.
    rewrite subword_sublist_eq in IH3.
    destruct IH3 as (pp & H3 & H4).
    apply sublist_eq in H4.
    subst pp.
    apply subword_intercalate; auto.
    apply Forall2_list_In_t_map.
    intros u Hu; apply H1.
    apply Forall2_length in H3.
    rewrite <- H3, map_length, map_length; omega.
  Qed.

  Hypothesis Heh : forall t, In_t t ll -> af_t (emb_tree_homeo R <# fun x => ~ t <eh x #>).  

  Fact vaz_rel_af_t : af_t (emb_tree_product R' <# stree vaz_ar #>).
  Proof.
    apply higman_stree_af_t with (k := S n).
    intros [ (?,(? & ?)) | ? ]; simpl; omega.
    intros i Hi.
    destruct (eq_nat_dec i n) as [ H | H ].
    
    (* case with i = n *)
    
    subst i.
    set (t u (_ : In_t u ll) := { ll | Forall (fun x => ~ u <eh x) ll }).
    set (S u (_ : In_t u ll) := subword (emb_tree_homeo R) <# Forall (fun x => ~ u <eh x) #>).
    generalize (af_t_idx_prod _ t S); intros H; spec_all H.

    unfold S.
    intros.
    apply afs_t_impl_af_t.
    apply higman_list_afs_t.
    intros a; destruct (Heh_dec x a); tauto.
    apply af_t_impl_afs_t.
    intros a; destruct (Heh_dec x a); tauto.
    apply Heh; auto.
    
    revert H.
    set (h (f : forall x Hx, @t x Hx) (b : { u | vaz_ar u = n } ) := 
      match b with 
        | exist (inr (_,g)) _  => forall x Hx, proj1_sig (f x Hx) = proj1_sig (g x Hx)
        |  _                   => False
      end).
    apply af_t_relmap with (f := h).
    
    intros ([ (?,(k,?)) | (?,f) ] & H); simpl in H.
    exfalso; omega.
    exists f; unfold h; auto.
    
    intros f1 f2 ([ (?,(k1,?)) | (?,g1) ] & H1) ([ (?,(k2,?)) | (?,g2) ] & H2); simpl; try tauto;
    intros H3 H4 H5; unfold rel_restr; simpl.
    intros ? ?; rewrite <- H3, <- H4; apply H5.
    
    (* case with i <> n *)
    
    apply in_af_t0.
    intros ([ (x1,(k1,Hk1)) | (x1,g1) ] & H1); simpl in H1; try (exfalso; omega).
    intros ([ (x2,(k2,Hk2)) | (x2,g2) ] & H2); simpl in H2; try (exfalso; omega).
    unfold rel_restr; simpl; auto.
  Qed.

End qtrees.
 
Section Vazsonyi_conjecture.

  Variable (X : Type) (R : X -> X -> Prop) (HR : forall x y, R x y).
  
  Let Heh : forall x y, { emb_tree_homeo R x y } + { ~ emb_tree_homeo R x y }.
  Proof.
    apply emb_tree_homeo_dec; left; auto.
  Defined.
  
  Theorem vazsonyi_af_t : af_t (emb_tree_homeo R).
  Proof.
    apply in_af_t1.
    intros t; induction t as [ x0 ll IH ].
    apply afs_t_af_t_rlift; auto.
    apply af_t_impl_afs_t; auto.
    intros x; destruct (Heh (in_tree x0 ll) x); tauto.
    
    case_eq (length ll).
    
    intros Hll.
    destruct ll; try discriminate Hll.
    apply in_af_t0.
    intros ([ x mm ] & Hx).
    contradict Hx.
    apply in_emb_tree_homeo_1; auto.
    apply subword_nil_true.
    
    intros n Hn.
    
    generalize (vaz_rel_af_t HR ll Hn); intros H; spec_all H.

    intros x Hx.
    apply afs_t_impl_af_t, af_t_rlift_afs_t; auto.
    revert H.
    apply af_t_relmap with (f := fun x y => @vaz_ev _ R n ll (proj1_sig x) = proj1_sig y).
    
    intros (y & Hy).
    destruct (@vaz_ev_surj _ R HR n x0 ll Hn y Hy) as (x & Hx1 & Hx2).
    exists (exist _ _ Hx2); simpl; auto.
    
    intros (x1 & Hx1) (x2 & Hx2) (y1 & Hy1) (y2 & Hy2); simpl.
    intros ? ?; subst y1 y2.
    unfold rel_restr; simpl.
    intros H; revert Hx1 Hx2.
    apply vaz_ev_emb; auto.
    revert H; apply emb_tree_product_emb_tree_homeo.
  Defined.
    
End Vazsonyi_conjecture.

Fact vazsonyi_conjecture : af_t (emb_tree_homeo (fun _ _ : unit => True)).
Proof.
  apply vazsonyi_af_t; auto.
Defined.

Check vazsonyi_conjecture.
Print Assumptions vazsonyi_conjecture.

(* Extraction "vazsonyi.ml" vazsonyi_conjecture. *)
